#!/usr/bin/env pytest

import re
import unittest

from fandango.constraints.expression import ExpressionConstraint
from fandango.language.grammar.grammar import Grammar
from fandango.language.search import ItemSearch, RuleSearch
from fandango.language.symbols import NonTerminal, Terminal
from fandango.language.tree import DerivationTree
from fandango.language.parse import parse
from .utils import RESOURCES_ROOT, run_command


class TestSlices(unittest.TestCase):
    TEST_DIR = RESOURCES_ROOT

    def test_startswith(self):
        command = [
            "fandango",
            "fuzz",
            "-f",
            str(self.TEST_DIR / "twodigits.fan"),
            "-n",
            "1",
            "-c",
            'str(<start>).startswith("6")',
            "--format=none",
            "--validate",
            "--random-seed",
            "426912",
            "--population-size",
            "10",
        ]
        out, err, code = run_command(command)
        self.assertEqual("", err)
        self.assertEqual("", out)
        self.assertEqual(0, code)

    def test_slice_0(self):
        command = [
            "fandango",
            "fuzz",
            "-f",
            str(self.TEST_DIR / "twodigits.fan"),
            "-n",
            "1",
            "-c",
            '<start>[0] == "6"',
            "--format=none",
            "--validate",
            "--random-seed",
            "426912",
            "--population-size",
            "10",
        ]
        out, err, code = run_command(command)
        self.assertEqual("", err)
        self.assertEqual("", out)
        self.assertEqual(0, code)

    def test_slice_0plus(self):
        command = [
            "fandango",
            "fuzz",
            "-f",
            str(self.TEST_DIR / "twodigits.fan"),
            "-n",
            "1",
            "-c",
            '<start>[0:][0] == "6"',
            "--format=none",
            "--validate",
            "--random-seed",
            "426912",
            "--population-size",
            "10",
        ]
        out, err, code = run_command(command)
        self.assertEqual("", err)
        self.assertEqual("", out)
        self.assertEqual(0, code)

    def test_slice_1plus(self):
        command = [
            "fandango",
            "fuzz",
            "-f",
            str(self.TEST_DIR / "twodigits.fan"),
            "-n",
            "1",
            "-c",
            '<start>[1:][0] == "6"',
            "--format=none",
            "--validate",
            "--random-seed",
            "426912",
            "--population-size",
            "10",
        ]
        out, err, code = run_command(command)
        self.assertEqual("", err)
        self.assertEqual("", out)
        self.assertEqual(0, code)

    def test_slice_str0plus1(self):
        command = [
            "fandango",
            "fuzz",
            "-f",
            str(self.TEST_DIR / "twodigits.fan"),
            "-n",
            "1",
            "-c",
            'str(<start>)[0:1] == "6"',
            "--format=none",
            "--validate",
            "--random-seed",
            "426912",
            "--population-size",
            "10",
        ]
        out, err, code = run_command(command)
        self.assertEqual("", err)
        self.assertEqual("", out)
        self.assertEqual(0, code)

    def test_slice_plus1(self):
        command = [
            "fandango",
            "fuzz",
            "-f",
            str(self.TEST_DIR / "twodigits.fan"),
            "-n",
            "1",
            "-c",
            '<start>[:1][0] == "6"',
            "--format=none",
            "--validate",
            "--random-seed",
            "426912",
            "--population-size",
            "10",
        ]
        out, err, code = run_command(command)
        self.assertEqual("", err)
        self.assertEqual("", out)
        self.assertEqual(0, code)

    def test_slice_plus(self):
        command = [
            "fandango",
            "fuzz",
            "-f",
            str(self.TEST_DIR / "twodigits.fan"),
            "-n",
            "1",
            "-c",
            '<start>[:][0] == "6"',
            "--format=none",
            "--validate",
            "--random-seed",
            "426912",
            "--population-size",
            "10",
        ]
        out, err, code = run_command(command)
        self.assertEqual("", err)
        self.assertEqual("", out)
        self.assertEqual(0, code)

    def test_slice_str0(self):
        command = [
            "fandango",
            "fuzz",
            "-f",
            str(self.TEST_DIR / "twodigits.fan"),
            "-n",
            "1",
            "-c",
            'str(<start>)[0] == "6"',
            "--format=none",
            "--validate",
            "--random-seed",
            "426912",
            "--population-size",
            "10",
        ]
        out, err, code = run_command(command)
        self.assertEqual("", err)
        self.assertEqual("", out)
        self.assertEqual(0, code)

    def test_slice_paren(self):
        command = [
            "fandango",
            "fuzz",
            "-f",
            str(self.TEST_DIR / "twodigits.fan"),
            "-n",
            "1",
            "-c",
            '(<start>)[0] == "6"',
            "--format=none",
            "--validate",
            "--random-seed",
            "426912",
            "--population-size",
            "10",
        ]
        out, err, code = run_command(command)
        self.assertEqual("", err)
        self.assertEqual("", out)
        self.assertEqual(0, code)


class TestExplicitSlices(unittest.TestCase):
    EXAMPLE = "\n".join(
        [
            "<start> ::= <a>",
            "<a> ::= <b> <c> <d>",
            '<b> ::= "a" | "b"',
            '<c> ::= "c"',
            '<d> ::= "d"',
            "",
            "where str(<a>[0]).endswith('b')",
        ]
    )

    VALID_EXAMPLE = DerivationTree(
        NonTerminal("<start>"),
        [
            DerivationTree(
                NonTerminal("<a>"),
                [
                    DerivationTree(
                        NonTerminal("<b>"),
                        [
                            DerivationTree(
                                Terminal("b"),
                            )
                        ],
                    ),
                    DerivationTree(
                        NonTerminal("<c>"),
                        [
                            DerivationTree(
                                Terminal("c"),
                            )
                        ],
                    ),
                    DerivationTree(
                        NonTerminal("<d>"),
                        [
                            DerivationTree(
                                NonTerminal("d"),
                            )
                        ],
                    ),
                ],
            )
        ],
    )

    INVALID_EXAMPLE = DerivationTree(
        NonTerminal("<start>"),
        [
            DerivationTree(
                NonTerminal("<a>"),
                [
                    DerivationTree(
                        NonTerminal("<b>"),
                        [
                            DerivationTree(
                                Terminal("a"),
                            )
                        ],
                    ),
                    DerivationTree(
                        NonTerminal("<c>"),
                        [
                            DerivationTree(
                                Terminal("c"),
                            )
                        ],
                    ),
                    DerivationTree(
                        NonTerminal("<d>"),
                        [
                            DerivationTree(
                                NonTerminal("d"),
                            )
                        ],
                    ),
                ],
            )
        ],
    )

    # Type annotations for instance attributes set in setUp
    GRAMMAR: Grammar
    CONSTRAINT: ExpressionConstraint

    @classmethod
    def setUpClass(cls):
        grammar, constraints = parse(cls.EXAMPLE, use_cache=False, use_stdlib=False)
        assert len(constraints) == 1
        assert isinstance(constraints[0], ExpressionConstraint)
        assert grammar is not None
        cls.GRAMMAR = grammar
        cls.CONSTRAINT = constraints[0]

    def test_parsed(self):
        self.assertTrue(
            self.CONSTRAINT.expression.endswith(".endswith('b')")
            or self.CONSTRAINT.expression.endswith('.endswith("b")')
        )
        expr = self.CONSTRAINT.expression
        pattern = r"___fandango.*___"
        re_search = re.search(pattern, expr)
        assert re_search is not None
        tmp_var = str(re_search.group(0))
        self.assertIn(tmp_var, self.CONSTRAINT.searches)
        search = self.CONSTRAINT.searches[tmp_var]
        assert search is not None
        assert isinstance(search, ItemSearch)
        base = search.base
        assert isinstance(base, RuleSearch)
        self.assertEqual(base.symbol, NonTerminal("<a>"))
        self.assertEqual(1, len(search.slices))
        self.assertEqual(0, search.slices[0])

    def test_valid(self):
        print(self.CONSTRAINT.format_as_spec())
        print(self.VALID_EXAMPLE.to_tree())
        self.assertTrue(self.CONSTRAINT.check(self.VALID_EXAMPLE))

    def test_invalid(self):
        self.assertFalse(self.CONSTRAINT.check(self.INVALID_EXAMPLE))



# fandango/evolution/algorithm.py
import enum
import itertools
import logging
import random
import time
import warnings
from collections.abc import Callable, Generator
from typing import Iterable, Optional, Union

from fandango.constraints.constraint import Constraint
from fandango.constraints.soft import SoftValue
from fandango.errors import FandangoFailedError, FandangoParseError, FandangoValueError
from fandango.evolution import GeneratorWithReturn
from fandango.evolution.adaptation import AdaptiveTuner
from fandango.evolution.crossover import CrossoverOperator, SimpleSubtreeCrossover
from fandango.evolution.evaluation import Evaluator
from fandango.evolution.mutation import MutationOperator, SimpleMutation
from fandango.evolution.population import IoPopulationManager, PopulationManager
from fandango.evolution.profiler import Profiler
from fandango.io import FandangoIO, FandangoParty
from fandango.io.packetforecaster import PacketForecaster
from fandango.io.packetparser import parse_next_remote_packet
from fandango.language.grammar import FuzzingMode
from fandango.language.grammar.grammar import Grammar
from fandango.language.tree import DerivationTree
from fandango.logger import (
    LOGGER,
    clear_visualization,
    log_message_transfer,
    print_exception,
    visualize_evaluation,
)


class LoggerLevel(enum.Enum):
    NOTSET = logging.NOTSET
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL


class Fandango:
    def __init__(
        self,
        grammar: Grammar,
        constraints: list[Union[Constraint, SoftValue]],
        population_size: int = 100,
        initial_population: Optional[list[Union[DerivationTree, str]]] = None,
        expected_fitness: float = 1.0,
        elitism_rate: float = 0.1,
        crossover_method: CrossoverOperator = SimpleSubtreeCrossover(),
        crossover_rate: float = 0.8,
        tournament_size: float = 0.1,
        mutation_method: MutationOperator = SimpleMutation(),
        mutation_rate: float = 0.2,
        destruction_rate: float = 0.0,
        logger_level: Optional[LoggerLevel] = None,
        warnings_are_errors: bool = False,
        best_effort: bool = False,
        random_seed: Optional[int] = None,
        start_symbol: str = "<start>",
        diversity_k: int = 5,
        diversity_weight: float = 1.0,
        max_repetition_rate: float = 0.5,
        max_repetitions: Optional[int] = None,
        max_nodes: int = 200,
        max_nodes_rate: float = 0.5,
        profiling: bool = False,
    ):
        if tournament_size > 1:
            raise FandangoValueError(
                f"Parameter tournament_size must be in range ]0, 1], but is {tournament_size}."
            )
        if random_seed is not None:
            random.seed(random_seed)
        if logger_level is not None:
            LOGGER.setLevel(logger_level.value)
        LOGGER.info("---------- Initializing FANDANGO algorithm ---------- ")

        self.grammar = grammar
        self.constraints = constraints
        self.population_size = population_size
        self.elitism_rate = elitism_rate
        self.destruction_rate = destruction_rate
        self.start_symbol = start_symbol
        self.tournament_size = tournament_size
        self.warnings_are_errors = warnings_are_errors
        self.best_effort = best_effort
        self.current_max_nodes = 50
        self.remote_response_timeout = 15.0

        # Instantiate managers
        if self.grammar.fuzzing_mode == FuzzingMode.IO:
            self.population_manager: PopulationManager = IoPopulationManager(
                grammar,
                start_symbol,
                warnings_are_errors,
            )
        else:
            self.population_manager = PopulationManager(
                grammar,
                start_symbol,
                warnings_are_errors,
            )
        self.evaluator = Evaluator(
            grammar,
            constraints,
            expected_fitness,
            diversity_k,
            diversity_weight,
            warnings_are_errors,
        )
        self.adaptive_tuner = AdaptiveTuner(
            mutation_rate,
            crossover_rate,
            grammar.get_max_repetition(),
            self.current_max_nodes,
            max_repetitions,
            max_repetition_rate,
            max_nodes,
            max_nodes_rate,
        )

        self.profiler = Profiler(enabled=profiling)

        self.crossover_operator = crossover_method
        self.mutation_method = mutation_method

        self.population = self._parse_and_deduplicate(population=initial_population)
        self._initial_solutions, self.evaluation = GeneratorWithReturn(
            self.evaluator.evaluate_population(self.population)
        ).collect()

        self.crossovers_made = 0
        self.fixes_made = 0
        self.mutations_made = 0
        self.time_taken = 0.0

    def _parse_and_deduplicate(
        self, population: Optional[list[Union[DerivationTree, str]]]
    ) -> list[DerivationTree]:
        """
        Parses and deduplicates the initial population along unique parse trees. If no initial population is provided, an empty list is returned.

        :param population: The initial population to parse and deduplicate.
        :return: A list of unique parse trees.
        """
        if population is None:
            return []
        LOGGER.info("Deduplicating the provided initial population...")
        unique_population: list[DerivationTree] = []
        unique_hashes: set[int] = set()
        for individual in population:
            if isinstance(individual, str):
                tree = self.grammar.parse(individual)
                if not tree:
                    position = self.grammar.max_position()
                    raise FandangoParseError(
                        message=f"Failed to parse initial individual{individual!r}",
                        position=position,
                    )
            elif isinstance(individual, DerivationTree):
                tree = individual
            else:
                raise TypeError("Initial individuals must be DerivationTree or String")
            PopulationManager.add_unique_individual(
                population=unique_population, candidate=tree, unique_set=unique_hashes
            )
        return unique_population

    def generate_initial_population(self) -> Generator[DerivationTree, None, None]:
        """
        Extends the population to the target size. Does not perform fixes.

        `.generate` will call this if necessary. If you don't know what you're doing, you probably don't need to call this.

        Since this is a generator, it will only do its job if the generator is actually used. Call `list(fandango.generate_initial_population())` to ensure the generator runs until the end.

        :return: A generator of DerivationTree objects, all of which are valid solutions to the grammar (or satisify the minimum fitness threshold).
        """
        LOGGER.info(
            f"Generating (additional) initial population (size: {self.population_size - len(self.population)})..."
        )
        st_time = time.time()

        with self.profiler.timer("initial_population") as timer:
            yield from self.population_manager.refill_population(
                current_population=self.population,
                eval_individual=self.evaluator.evaluate_individual,
                max_nodes=self.current_max_nodes,
                target_population_size=self.population_size,
            )

            timer.increment(len(self.population))

        LOGGER.info(
            f"Initial population generated in {time.time() - st_time:.2f} seconds"
        )

        # Evaluate initial population
        with self.profiler.timer("evaluate_population", increment=self.population):
            self.evaluation = yield from self.evaluator.evaluate_population(
                self.population
            )

    def _perform_selection(self) -> tuple[list[DerivationTree], set[int]]:
        """
        Performs selection of the elites from the population.

        :return: A tuple containing the new population and the set of unique hashes of the individuals in the new population.
        """
        # defer increment until data is available
        with self.profiler.timer("select_elites") as timer:
            new_population = self.evaluator.select_elites(
                self.evaluation,
                self.elitism_rate,
                self.population_size,
            )
            timer.increment(len(new_population))

        unique_hashes = {hash(ind) for ind in new_population}
        return new_population, unique_hashes

    def _perform_crossover(
        self, new_population: list[DerivationTree], unique_hashes: set[int]
    ) -> Generator[DerivationTree, None, None]:
        """
        Performs crossover of the population.

        :param new_population: The new population to perform crossover on.
        :param unique_hashes: The set of unique hashes of the individuals in the new population.
        """
        if len(self.evaluation) < 2:
            return None
        try:
            with self.profiler.timer("tournament_selection", increment=2):
                parent1, parent2 = self.evaluator.tournament_selection(
                    evaluation=self.evaluation,
                    tournament_size=max(
                        2, int(self.population_size * self.tournament_size)
                    ),
                )

            with self.profiler.timer("crossover", increment=2):
                child1, child2 = self.crossover_operator.crossover(
                    self.grammar, parent1, parent2
                )

            PopulationManager.add_unique_individual(
                new_population, child1, unique_hashes
            )
            yield from self.evaluator.evaluate_individual(child1)

            count = len(new_population)
            with self.profiler.timer("filling") as timer:
                if len(new_population) < self.population_size:
                    PopulationManager.add_unique_individual(
                        new_population, child2, unique_hashes
                    )
                yield from self.evaluator.evaluate_individual(child2)
                timer.increment(len(new_population) - count)
            self.crossovers_made += 2
        except Exception as e:
            print_exception(e, "Error during crossover")

    def _perform_mutation(
        self, new_population: list[DerivationTree]
    ) -> Generator[DerivationTree, None, None]:
        """
        Performs mutation of the population.

        :param new_population: The new population to perform mutation on.
        """
        mutation_pool = self.evaluator.compute_mutation_pool(new_population)
        mutated_population = []
        for individual in mutation_pool:
            if random.random() < self.adaptive_tuner.mutation_rate:
                try:
                    with self.profiler.timer("mutation", increment=1):
                        mutated_individual = yield from self.mutation_method.mutate(
                            individual,
                            self.grammar,
                            self.evaluator.evaluate_individual,
                        )
                    mutated_population.append(mutated_individual)
                    self.mutations_made += 1
                except Exception as e:
                    print_exception(e, "Error during mutation")
                    mutated_population.append(individual)
            else:
                mutated_population.append(individual)
        new_population.extend(mutated_population)

    def _perform_destruction(
        self, new_population: list[DerivationTree]
    ) -> list[DerivationTree]:
        """
        Randomly destroys a portion of the population.

        :param new_population: The new population to perform destruction on.
        :return: The new population after destruction.
        """
        LOGGER.debug(f"Destroying {self.destruction_rate * 100:.2f}% of the population")
        random.shuffle(new_population)
        return new_population[: int(self.population_size * (1 - self.destruction_rate))]

    def evolve(
        self,
        max_generations: Optional[int] = None,
        desired_solutions: Optional[int] = None,
        solution_callback: Callable[[DerivationTree, int], None] = lambda _a, _b: None,
    ) -> list[DerivationTree]:
        """
        Evolves the population of the grammar.

        If both max_generations and desired_solutions are provided, the generation will run until either the maximum number of generations is reached or the desired number of solutions is found. If neither is provided, the generation will run indefinitely.

        TODO: go into more details about Fandango IO mode.

        :param max_generations: The maximum number of generations to evolve.
        :param desired_solutions: The number of solutions to evolve.
        :param solution_callback: A callback function to be called for each solution.
        :return: A list of DerivationTree objects, all of which are valid solutions to the grammar (or satisify the minimum fitness threshold). The function may run indefinitely if neither max_generations nor desired_solutions are provided.
        """
        warnings.warn("Use .generate instead", DeprecationWarning)
        if self.grammar.fuzzing_mode == FuzzingMode.COMPLETE:
            return self._evolve_single(
                max_generations, desired_solutions, solution_callback
            )
        elif self.grammar.fuzzing_mode == FuzzingMode.IO:
            return self.evolve_io(max_generations)
        else:
            raise FandangoValueError(f"Invalid mode: {self.grammar.fuzzing_mode}")

    def evolve_io(self, max_generations: Optional[int] = None) -> list[DerivationTree]:
        warnings.warn("Use .generate instead", DeprecationWarning)
        return list(self._generate_io(max_generations=max_generations))

    def generate(
        self,
        max_generations: Optional[int] = None,
        mode: FuzzingMode = FuzzingMode.COMPLETE,
    ) -> Generator[DerivationTree, None, None]:
        match mode:
            case FuzzingMode.COMPLETE:
                yield from self._generate_simple(max_generations=max_generations)
            case FuzzingMode.IO:
                yield from self._generate_io(max_generations=max_generations)
            case _:
                raise RuntimeError(f"Fuzzing Mode {mode} is not implemented")

    def _generate_simple(
        self, max_generations: Optional[int] = None
    ) -> Generator[DerivationTree, None, None]:
        """
        Generates solutions for the grammar.

        :param max_generations: The maximum number of generations to generate. If None, the generation will run indefinitely.
        :return: A generator of DerivationTree objects, all of which are valid solutions to the grammar (or satisify the minimum fitness threshold).
        """
        yield from self._initial_solutions
        self._initial_solutions.clear()

        if len(self.population) < self.population_size:
            yield from self.generate_initial_population()

        prev_best_fitness = 0.0
        generation = 0

        while True:
            if max_generations is not None and generation >= max_generations:
                break
            generation += 1

            avg_fitness = sum(e[1] for e in self.evaluation) / self.population_size

            LOGGER.info(f"Generation {generation} - Average Fitness: {avg_fitness:.2f}")

            # Selection
            new_population, unique_hashes = self._perform_selection()

            # Crossover
            while (
                len(new_population) < self.population_size
                and random.random() < self.adaptive_tuner.crossover_rate
            ):
                yield from self._perform_crossover(new_population, unique_hashes)

            # Truncate if necessary
            if len(new_population) > self.population_size:
                new_population = new_population[: self.population_size]

            # Mutation
            yield from self._perform_mutation(new_population)

            # Destruction
            if self.destruction_rate > 0:
                new_population = self._perform_destruction(new_population)

            # Ensure Uniqueness & Fill Population
            new_population = list(set(new_population))
            yield from self.population_manager.refill_population(
                new_population,
                self.evaluator.evaluate_individual,
                self.current_max_nodes,
                self.population_size,
            )

            self.population = []
            for ind in new_population:
                _fitness, failing_trees = yield from self.evaluator.evaluate_individual(
                    ind
                )
                ind, num_fixes = self.population_manager.fix_individual(
                    ind, failing_trees
                )
                self.population.append(ind)
                self.fixes_made += num_fixes

            # For soft constraints, the normalized fitness may change over time as we observe more inputs.
            # Hence, we periodically flush the fitness cache to re-evaluate the population if the grammar contains soft constraints.
            self.evaluator.flush_fitness_cache()

            with self.profiler.timer("evaluate_population", increment=self.population):
                self.evaluation = yield from self.evaluator.evaluate_population(
                    self.population
                )
                # Keep only the fittest individuals
                self.evaluation = sorted(
                    self.evaluation, key=lambda x: x[1], reverse=True
                )[: self.population_size]

            current_best_fitness = max(e[1] for e in self.evaluation)
            current_max_repetitions = self.grammar.get_max_repetition()
            self.adaptive_tuner.update_parameters(
                generation,
                prev_best_fitness,
                current_best_fitness,
                self.population,
                self.evaluator,
                current_max_repetitions,
            )

            if self.adaptive_tuner.current_max_repetition > current_max_repetitions:
                self.grammar.set_max_repetition(
                    self.adaptive_tuner.current_max_repetition
                )

            self.current_max_nodes = self.adaptive_tuner.current_max_nodes

            prev_best_fitness = current_best_fitness

            self.adaptive_tuner.log_generation_statistics(
                generation, self.evaluation, self.population, self.evaluator
            )
            visualize_evaluation(generation, max_generations, self.evaluation)
        clear_visualization()
        self.log_statistics()

    def _generate_io(
        self, max_generations: Optional[int] = None
    ) -> Generator[DerivationTree, None, None]:
        if len(self.population) < self.population_size:
            list(
                self.generate_initial_population()
            )  # ensure the generator runs until the end

        spec_env_global, _ = self.grammar.get_spec_env()
        io_instance: FandangoIO = spec_env_global["FandangoIO"].instance()
        history_tree: DerivationTree = random.choice(self.population)
        forecaster = PacketForecaster(self.grammar)

        while True:
            forecast = forecaster.predict(history_tree)

            if len(forecast.get_msg_parties()) == 0:
                if len(history_tree.protocol_msgs()) == 0:
                    raise FandangoFailedError("Could not forecast next packet")
                yield history_tree
                return
                # TODO: Reset for next iteration

            msg_parties = list(
                filter(
                    lambda x: io_instance.parties[x].is_fuzzer_controlled(),
                    forecast.get_msg_parties(),
                )
            )
            if len(msg_parties) != 0 and not io_instance.received_msg():
                fuzzable_packets = []
                for party in msg_parties:
                    fuzzable_packets.extend(forecast[party].nt_to_packet.values())
                assert isinstance(self.population_manager, IoPopulationManager)
                self.population_manager.fuzzable_packets = fuzzable_packets

                self.population.clear()
                solutions = list(
                    self.population_manager.refill_population(
                        current_population=self.population,
                        eval_individual=self.evaluator.evaluate_individual,
                        max_nodes=self.current_max_nodes,
                        target_population_size=self.population_size,
                    )
                )
                if not solutions:
                    solutions, self.evaluation = GeneratorWithReturn(
                        self.evaluator.evaluate_population(self.population)
                    ).collect()

                if not solutions:
                    try:
                        evolve_result = next(
                            self.generate(max_generations=max_generations)
                        )
                    except StopIteration:
                        nonterminals_str = " | ".join(
                            map(lambda x: str(x.node.symbol), fuzzable_packets)
                        )
                        raise FandangoFailedError(
                            f"Couldn't find solution for any packet: {nonterminals_str}"
                        )
                    next_tree = evolve_result
                else:
                    next_tree = solutions[0]
                if io_instance.received_msg():
                    # Abort if we received a message during fuzzing
                    continue
                new_packet = next_tree.protocol_msgs()[-1]
                if (
                    new_packet.recipient is None
                    or not io_instance.parties[
                        new_packet.recipient
                    ].is_fuzzer_controlled()
                ):
                    io_instance.transmit(
                        new_packet.sender, new_packet.recipient, new_packet.msg
                    )
                    log_message_transfer(
                        new_packet.sender, new_packet.recipient, new_packet.msg, True
                    )
                history_tree = next_tree
            else:
                wait_start = time.time()
                while not io_instance.received_msg():
                    if time.time() - wait_start > self.remote_response_timeout:
                        raise FandangoFailedError(
                            f"Timed out while waiting for message from remote party. Expected message from party: {', '.join(forecast.get_msg_parties())}"
                        )
                    time.sleep(0.025)
                forecast, packet_tree = parse_next_remote_packet(
                    self.grammar, forecast, io_instance
                )
                log_message_transfer(
                    packet_tree.sender,
                    packet_tree.recipient,
                    packet_tree,
                    False,
                )

                hookin_option = next(iter(forecast.paths))
                history_tree = hookin_option.tree
                history_tree.append(hookin_option.path[1:], packet_tree)
                solutions, (fitness, _failing_trees) = GeneratorWithReturn(
                    self.evaluator.evaluate_individual(history_tree)
                ).collect()
                if fitness < 0.99:
                    raise FandangoParseError(
                        "Remote response does not match constraints"
                    )
            history_tree.set_all_read_only(True)

    @property
    def average_population_fitness(self) -> float:
        return sum(e[1] for e in self.evaluation) / self.population_size

    def log_statistics(self) -> None:
        LOGGER.debug("---------- FANDANGO statistics ----------")
        LOGGER.info(
            f"Average fitness of population: {self.average_population_fitness:.2f}"
        )
        LOGGER.debug(f"Fixes made: {self.fixes_made}")
        LOGGER.debug(f"Fitness checks: {self.evaluator.get_fitness_check_count()}")
        LOGGER.debug(f"Crossovers made: {self.crossovers_made}")
        LOGGER.debug(f"Mutations made: {self.mutations_made}")
        self.profiler.log_results()

    def _evolve_single(
        self,
        max_generations: Optional[int] = None,
        desired_solutions: Optional[int] = None,
        solution_callback: Callable[[DerivationTree, int], None] = lambda _a, _b: None,
    ) -> list[DerivationTree]:
        LOGGER.info("---------- Starting evolution ----------")

        solutions: list[DerivationTree] = []

        start_time = time.time()
        gen: Iterable[DerivationTree] = self.generate(max_generations)
        if desired_solutions is not None:
            gen = itertools.islice(gen, desired_solutions)

        for solution in gen:
            solutions.append(solution)
            solution_callback(solution, len(solutions))
        LOGGER.info(f"Time taken: {(time.time() - start_time):.2f} seconds")

        return solutions

    def msg_parties(self) -> list[FandangoParty]:
        """
        :return: A list of all parties in the grammar.
        """
        spec_env_global, _ = self.grammar.get_spec_env()
        io_instance: FandangoIO = spec_env_global["FandangoIO"].instance()
        return list(io_instance.parties.values())



from fandango.language.symbols import Symbol, SymbolType
from fandango.language.tree_value import TreeValueType


class NonTerminal(Symbol):
    def __init__(self, symbol: str) -> None:
        assert isinstance(symbol, str)
        super().__init__(symbol, SymbolType.NON_TERMINAL)

    def name(self) -> str:
        """
        Return the name of the non-terminal symbol.
        """
        return str(self._value)

    def __hash__(self) -> int:
        return hash((self._value, self._type))

    def format_as_spec(self) -> str:
        return self.name()



from copy import deepcopy
from typing import Optional
from collections.abc import Generator

from fandango.language.grammar import ParsingMode
from fandango.language.grammar.parser.iterative_parser import IterativeParser
from fandango.language.grammar.nodes.node import Node
from fandango.language.symbols.non_terminal import NonTerminal
from fandango.language.tree import DerivationTree


class Parser:

    def __init__(self, grammar_rules: dict[NonTerminal, Node]):
        self._iter_parser = IterativeParser(grammar_rules)
        self._cache: dict[
            tuple[
                str | bytes,
                NonTerminal,
                ParsingMode,
                Optional[DerivationTree],
            ],
            list[DerivationTree],
        ] = {}

    def _parse_forest(
        self,
        word: str | bytes,
        start: str | NonTerminal = "<start>",
        *,
        mode: ParsingMode = ParsingMode.COMPLETE,
        hookin_parent: Optional[DerivationTree] = None,
        starter_bit=-1,
    ):
        """
        Parse a forest of input trees from `word`.
        `start` is the start symbol (default: `<start>`).
        if `allow_incomplete` is True, the function will return trees even if the input ends prematurely.
        """
        self._iter_parser.new_parse(start, mode, hookin_parent, starter_bit)
        yield from self._iter_parser.consume(word)

    def parse_forest(
        self,
        word: str | bytes | int | DerivationTree,
        start: str | NonTerminal = "<start>",
        mode: ParsingMode = ParsingMode.COMPLETE,
        hookin_parent: Optional[DerivationTree] = None,
        include_controlflow: bool = False,
    ) -> Generator[Optional[DerivationTree], None, None]:
        """
        Yield multiple parse alternatives, using a cache.
        """
        starter_bit = -1
        if isinstance(word, DerivationTree):
            if word.contains_bits():
                starter_bit = (word.count_terminals() - 1) % 8
            if word.should_be_serialized_to_bytes():
                bit_string = word.to_bits()
                word = int(bit_string, 2).to_bytes(
                    (len(bit_string) + 7) // 8, byteorder="big"
                )
            else:
                word = word.to_string()
        if isinstance(word, int):
            word = str(word)
        assert isinstance(word, (str, bytes)), type(word)

        if isinstance(start, str):
            start = NonTerminal(start)

        cache_key = (word, start, mode, hookin_parent)
        forest: list[DerivationTree]
        if cache_key in self._cache:
            forest = self._cache[cache_key]
            for tree in forest:
                tree = deepcopy(tree)
                if not include_controlflow:
                    yield self.collapse(tree)
            return

        for tree in self._parse_forest(
            word,
            start,
            mode=mode,
            hookin_parent=hookin_parent,
            starter_bit=starter_bit,
        ):
            tree = self._iter_parser.to_derivation_tree(tree)
            if cache_key in self._cache:
                self._cache[cache_key].append(tree)
            else:
                self._cache[cache_key] = [tree]
            yield self.collapse(tree) if not include_controlflow else tree

    def parse_multiple(
        self,
        word: str | bytes | int | DerivationTree,
        start: str | NonTerminal = "<start>",
        mode: ParsingMode = ParsingMode.COMPLETE,
        hookin_parent: Optional[DerivationTree] = None,
        include_controlflow: bool = False,
    ):
        """
        Yield multiple parse alternatives,
        even for incomplete inputs
        """
        return self.parse_forest(
            word,
            start,
            mode=mode,
            hookin_parent=hookin_parent,
            include_controlflow=include_controlflow,
        )

    def parse(
        self,
        word: str | bytes | int | DerivationTree,
        start: str | NonTerminal = "<start>",
        mode: ParsingMode = ParsingMode.COMPLETE,
        hookin_parent: Optional[DerivationTree] = None,
        include_controlflow: bool = False,
    ):
        """
        Return the first parse alternative,
        or `None` if no parse is possible
        """
        tree_gen = self.parse_multiple(
            word,
            start=start,
            mode=mode,
            hookin_parent=hookin_parent,
            include_controlflow=include_controlflow,
        )
        return next(tree_gen, None)

    def collapse(self, tree: Optional[DerivationTree]) -> Optional[DerivationTree]:
        return self._iter_parser.collapse(tree)



import ast
import hashlib
import os
import platform
import re
import sys
import time
import shutil

from copy import deepcopy
from io import StringIO
from pathlib import Path
from types import NoneType
from typing import IO, Optional
import warnings

import cachedir_tag
import dill as pickle
from antlr4 import CommonTokenStream, InputStream
from antlr4.error.ErrorListener import ErrorListener
from antlr4.tree.Tree import ParseTree
from xdg_base_dirs import xdg_cache_home, xdg_data_dirs, xdg_data_home

import fandango
from fandango.errors import FandangoSyntaxError, FandangoValueError
from fandango.constraints import predicates
from fandango.constraints.constraint import Constraint
from fandango.constraints.soft import SoftValue
from fandango.io import FandangoIO, FandangoParty
from fandango.language.convert import (
    ConstraintProcessor,
    FandangoSplitter,
    GrammarProcessor,
    PythonProcessor,
)
from fandango.language.grammar import FuzzingMode, closest_match
from fandango.language.grammar.grammar import Grammar
from fandango.language.grammar.node_visitors.message_nesting_detector import (
    MessageNestingDetector,
)
from fandango.language.grammar.node_visitors.node_replacer import NodeReplacer
from fandango.language.grammar.node_visitors.packet_truncator import PacketTruncator
from fandango.language.grammar.node_visitors.symbol_finder import SymbolFinder
from fandango.language.grammar.nodes.node import Node, NodeType
from fandango.language.grammar.nodes.non_terminal import NonTerminalNode
from fandango.language.grammar.nodes.repetition import Option, Plus, Repetition, Star
from fandango.language.grammar.nodes.terminal import TerminalNode
from fandango.language.parser import sa_fandango
from fandango.language.parser.FandangoLexer import FandangoLexer
from fandango.language.parser.FandangoParser import FandangoParser
from fandango.language.search import DescendantAttributeSearch, ItemSearch
from fandango.language.stdlib import stdlib
from fandango.language.symbols import NonTerminal, Symbol
from fandango.language.tree_value import TreeValueType
from fandango.logger import LOGGER, print_exception


class PythonAntlrErrorListener(ErrorListener):
    """This is invoked from ANTLR when a syntax error is encountered"""

    def __init__(self, filename=None):
        self.filename = filename
        super().__init__()

    def syntaxError(
        self, recognizer, offendingSymbol, line: int, column: int, msg: str, e
    ):
        raise FandangoSyntaxError(
            f"{self.filename!r}, line {line}, column {column}: {msg}"
        )


class SpeedyAntlrErrorListener(sa_fandango.SA_ErrorListener):
    """This is invoked from the speedy ANTLR parser when a syntax error is encountered"""

    def __init__(self, filename=None):
        self.filename = filename
        super().__init__()

    def syntaxError(
        self,
        input_stream,
        offending_symbol,
        char_index: int,
        line: int,
        column: int,
        msg,
    ):
        raise FandangoSyntaxError(
            f"{self.filename!r}, line {line}, column {column}: {msg}"
        )


### Including Files

# Some global variables for `include()`, bwlow

# The current file name, for error messages
CURRENT_FILENAME: str = "<undefined>"

# The list of directories to search for include files
INCLUDES: list[str] = []

# The list of files to parse, with their include depth.
# An include depth of 0 means the file was given as input_.
# A higher include depth means the file was included from another file;
# hence its grammar and constraints should be processed _before_ the current file.
# Format: (file_name, file_contents, include_depth)
FILES_TO_PARSE: list[tuple[str, str, int]] = []

# The current include depth
INCLUDE_DEPTH: int = 0


def include(file_to_be_included: str):
    """
    Include FILE_TO_BE_INCLUDED in the current context.
    This function is invoked from .fan files.
    """
    global FILES_TO_PARSE
    global CURRENT_FILENAME
    global INCLUDE_DEPTH

    path = os.path.dirname(CURRENT_FILENAME)
    if not path:
        # If the current file has no path, use the current directory
        path = "."
    if INCLUDES:
        path += ":" + ":".join(INCLUDES)
    if os.environ.get("FANDANGO_PATH"):
        path += ":" + os.environ["FANDANGO_PATH"]
    dirs = [Path(dir) for dir in path.split(":")]

    if platform.system() == "Darwin":
        dirs += [Path.home() / "Library" / "Fandango"]  # ~/Library/Fandango
        dirs += [Path("/Library/Fandango")]  # /Library/Fandango

    dirs += [xdg_data_home() / "fandango"]  # sth like ~/.local/share/fandango
    dirs += [
        dir / "fandango" for dir in xdg_data_dirs()
    ]  # sth like /usr/local/share/fandango

    for dir in dirs:
        full_file_name = dir / file_to_be_included
        if not os.path.exists(full_file_name):
            continue
        with open(full_file_name, "r") as full_file:
            LOGGER.debug(f"{CURRENT_FILENAME}: including {full_file_name}")

            INCLUDE_DEPTH += (
                1  # Will be lowered when the included file is done processing
            )
            FILES_TO_PARSE.append((full_file.name, full_file.read(), INCLUDE_DEPTH))
        return

    raise FileNotFoundError(
        f"{CURRENT_FILENAME}: {file_to_be_included!r} not found in {':'.join(str(dir) for dir in dirs)}"
    )


### Parsing


class FandangoSpec:
    """
    Helper class to pickle and unpickle parsed Fandango specifications.
    This is necessary because the ANTLR4 parse trees cannot be pickled,
    so we pickle the code text, grammar, and constraints instead.
    """

    GLOBALS = predicates.__dict__
    GLOBALS.update({"include": include})
    LOCALS = None  # Must be None to ensure top-level imports

    def __init__(
        self,
        tree: ParseTree,
        fan_contents: str,
        lazy: bool = False,
        filename: str = "<input_>",
        max_repetitions: int = 5,
    ):
        self.version = fandango.version()
        self.fan_contents = fan_contents
        self.global_vars = self.GLOBALS.copy()
        self.local_vars = self.LOCALS
        self.lazy = lazy

        LOGGER.debug(f"{filename}: extracting code")
        splitter = FandangoSplitter()
        splitter.visit(tree)
        python_processor = PythonProcessor()
        code_tree = python_processor.get_code(splitter.python_code)
        ast.fix_missing_locations(code_tree)
        self.code_text = ast.unparse(code_tree)

        LOGGER.debug(f"{filename}: code text:\n{self.code_text}")

        LOGGER.debug(f"{filename}: running code")
        self.run_code(filename=filename)

        LOGGER.debug(f"{filename}: extracting grammar")
        grammar_processor = GrammarProcessor(
            splitter.grammar_settings,
            local_variables=self.local_vars,
            global_variables=self.global_vars,
            id_prefix="{0:x}".format(abs(hash(filename))),
            max_repetitions=max_repetitions,
        )
        self.grammar = grammar_processor.get_grammar(splitter.productions, prime=False)

        LOGGER.debug(f"{filename}: extracting constraints")
        constraint_processor = ConstraintProcessor(
            self.grammar,
            local_variables=self.local_vars,
            global_variables=self.global_vars,
            lazy=self.lazy,
        )
        self.constraints: list[Constraint | SoftValue] = (
            constraint_processor.get_constraints(splitter.constraints)
        )
        self.constraints.extend(grammar_processor.repetition_constraints)

    def run_code(self, filename: str = "<input_>"):
        global CURRENT_FILENAME
        CURRENT_FILENAME = filename

        # Ensure the directory of the file is in the path
        dirname = os.path.dirname(filename)
        if dirname not in sys.path:
            sys.path.append(dirname)

        # Set up environment as if this were a top-level script
        self.global_vars.update(
            {
                "__name__": "__main__",
                "__file__": filename,
                "__package__": None,
                "__spec__": None,
            }
        )
        exec(self.code_text, self.global_vars, self.local_vars)

    def __repr__(self):
        s = self.code_text
        if s:
            s += "\n\n"
        s += str(self.grammar) + "\n"
        if self.constraints:
            s += "\n"
        s += "\n".join(
            "where " + constraint.format_as_spec() for constraint in self.constraints
        )
        return s


def cache_dir() -> Path:
    """Return the parser cache directory"""
    CACHE_DIR = xdg_cache_home() / "fandango"
    if platform.system() == "Darwin":
        cache_path = Path.home() / "Library" / "Caches"
        if os.path.exists(cache_path):
            CACHE_DIR = cache_path / "Fandango"
    return CACHE_DIR


def clear_cache() -> None:
    """Clear the Fandango parser cache"""
    CACHE_DIR = cache_dir()
    if os.path.exists(CACHE_DIR):
        shutil.rmtree(CACHE_DIR, ignore_errors=True)


def parse_spec(
    fan_contents: str,
    *,
    filename: str = "<input_>",
    use_cache: bool = True,
    lazy: bool = False,
    parties: list[str] | None = None,
    max_repetitions: int = 5,
) -> FandangoSpec:
    """
    Parse given content into a grammar and constraints.
    This is a helper function; use `parse()` as the main entry point.
    :param fan_contents: Fandango specification text
    :param filename: The file name of the content (for error messages)
    :param use_cache: If True (default), cache parsing results
    :param parties: If given, list of parties to consider in the grammar
    :param lazy: If True, the constraints are evaluated lazily
    :return: A FandangoSpec object containing the parsed grammar, constraints, and code text.
    """
    spec: Optional[FandangoSpec] = None
    from_cache = False

    CACHE_DIR = cache_dir()
    if use_cache:
        if not os.path.exists(CACHE_DIR):
            os.makedirs(CACHE_DIR, mode=0o700)
            cachedir_tag.tag(CACHE_DIR, application="Fandango")

        # Keep separate hashes for different Fandango and Python versions
        hash_contents = fan_contents + fandango.version() + "-" + sys.version
        hash = hashlib.sha256(hash_contents.encode()).hexdigest()
        pickle_file = CACHE_DIR / (hash + ".pickle")

        if os.path.exists(pickle_file):
            try:
                with open(pickle_file, "rb") as fp:
                    LOGGER.info(f"{filename}: loading cached spec from {pickle_file}")
                    start_time = time.time()
                    with warnings.catch_warnings():
                        warnings.simplefilter(
                            "ignore", DeprecationWarning
                        )  # for some reason, unpickling triggers the deprecation warnings in __getattr__ of DerivationTree and TreeValue
                        spec = pickle.load(fp)
                    assert spec is not None
                    LOGGER.debug(f"Cached spec version: {spec.version}")
                    if spec.fan_contents != fan_contents:
                        error = FandangoValueError(
                            "Hash collision (If you get this, you'll be real famous)"
                        )
                        raise error

                    LOGGER.debug(
                        f"{filename}: loaded from cache in {time.time() - start_time:.2f} seconds"
                    )
                    from_cache = True
            except Exception as exc:

                LOGGER.debug(type(exc).__name__ + ":" + str(exc))

    if spec:
        LOGGER.debug(f"{filename}: running code")
        try:
            spec.run_code(filename=filename)
        except Exception as exc:
            # In case the error has anything to do with caching, play it safe
            LOGGER.debug(f"Cached spec failed; removing {pickle_file}")
            os.remove(pickle_file)
            raise exc

    error_listener: SpeedyAntlrErrorListener | PythonAntlrErrorListener
    if not spec:
        if fandango.Fandango.parser != "legacy":
            if fandango.Fandango.parser == "cpp":
                sa_fandango.USE_CPP_IMPLEMENTATION = True
                try:
                    from .parser import sa_fandango_cpp_parser  # type: ignore[attr-defined]  # noqa: F401
                except ImportError:
                    raise ImportError(
                        "Requested C++ parser not available. "
                        "Check your installation "
                        "or use '--parser=python'"
                    )
            elif fandango.Fandango.parser == "python":
                sa_fandango.USE_CPP_IMPLEMENTATION = False
            elif fandango.Fandango.parser == "auto":
                pass  # let sa_fandango decide

            if sa_fandango.USE_CPP_IMPLEMENTATION:
                LOGGER.debug(f"{filename}: setting up C++ .fan parser")
            else:
                LOGGER.debug(f"{filename}: setting up Python .fan parser")

            input_stream = InputStream(fan_contents)
            error_listener = SpeedyAntlrErrorListener(filename)

            # Invoke the Speedy ANTLR parser
            LOGGER.debug(f"{filename}: parsing .fan content")
            start_time = time.time()
            tree = sa_fandango.parse(input_stream, "fandango", error_listener)
            LOGGER.debug(
                f"{filename}: parsed in {time.time() - start_time:.2f} seconds"
            )

        else:  # legacy parser
            LOGGER.debug(f"{filename}: setting up legacy .fan parser")
            input_stream = InputStream(fan_contents)
            error_listener = PythonAntlrErrorListener(filename)

            lexer = FandangoLexer(input_stream)
            lexer.removeErrorListeners()
            lexer.addErrorListener(error_listener)

            token_stream = CommonTokenStream(lexer)
            parser = FandangoParser(token_stream)
            parser.removeErrorListeners()
            parser.addErrorListener(error_listener)

            # Invoke the ANTLR parser
            LOGGER.debug(f"{filename}: parsing .fan content")
            start_time = time.time()
            tree = parser.fandango()

            LOGGER.debug(
                f"{filename}: parsed in {time.time() - start_time:.2f} seconds"
            )

        LOGGER.debug(f"{filename}: splitting content")
        spec = FandangoSpec(
            tree, fan_contents, lazy, filename=filename, max_repetitions=max_repetitions
        )

    assert spec is not None

    if use_cache and not from_cache:
        try:
            with open(pickle_file, "wb") as fp:
                LOGGER.info(f"{filename}: saving spec to cache {pickle_file}")
                pickle.dump(spec, fp)
        except Exception as e:
            print_exception(e)
            try:
                os.remove(pickle_file)  # might be inconsistent
            except Exception:
                pass

    if parties:
        slice_parties(spec.grammar, parties)

    LOGGER.debug(f"{filename}: parsing complete")
    return spec


# Legacy interface
def parse_content(*args, **kwargs) -> tuple[Grammar, list[Constraint | SoftValue]]:
    spec = parse_spec(*args, **kwargs)
    return spec.grammar, spec.constraints


# Save the set of symbols used in the standard library and imported grammars
USED_SYMBOLS: set[str] = set()

# Save the standard library grammar and constraints
STDLIB_GRAMMAR: Optional[Grammar] = None
STDLIB_CONSTRAINTS: Optional[list[Constraint | SoftValue]] = None


def parse(
    fan_files: str | IO | list[str | IO],
    constraints: Optional[list[str]] = None,
    *,
    use_cache: bool = True,
    use_stdlib: bool = True,
    check: bool = True,
    lazy: bool = False,
    given_grammars: list[Grammar] = [],
    start_symbol: Optional[str] = None,
    includes: Optional[list[str]] = [],
    parties: Optional[list[str]] = None,
    max_repetitions: int = 5,
) -> tuple[Optional[Grammar], list[Constraint | SoftValue]]:
    """
    Parse .fan content, handling multiple files, standard library, and includes.
    :param fan_files: One (open) .fan file, one string, or a list of these
    :param constraints: List of constraints (as strings); default: []
    :param use_cache: If True (default), cache parsing results
    :param use_stdlib: If True (default), use the standard library
    :param check: If True (default), the constraints are checked for consistency
    :param lazy: If True, the constraints are evaluated lazily
    :param given_grammars: Grammars to use in addition to the standard library
    :param start_symbol: The grammar start symbol (default: "<start>")
    :param includes: A list of directories to search for include files; default: []
    :param parties: If given, list of parties to consider in the grammar
    :param max_repetitions: The maximal number of repetitions
    :return: A tuple of the grammar and constraints
    """

    if not isinstance(fan_files, list):
        fan_files = [fan_files]

    if not fan_files and not constraints:
        return None, []

    if constraints is None:
        constraints = []

    if includes is None:
        includes = []

    if start_symbol is None:
        start_symbol = "<start>"

    global STDLIB_SYMBOLS, STDLIB_GRAMMAR, STDLIB_CONSTRAINTS
    if use_stdlib and STDLIB_GRAMMAR is None:
        LOGGER.debug("Reading standard library")
        STDLIB_GRAMMAR, STDLIB_CONSTRAINTS = parse_content(
            stdlib,
            filename="<stdlib>",
            use_cache=use_cache,
            max_repetitions=max_repetitions,
        )

    global USED_SYMBOLS
    USED_SYMBOLS = set()
    if use_stdlib:
        assert STDLIB_GRAMMAR is not None
        for symbol in STDLIB_GRAMMAR.rules.keys():
            # Do not complain about unused symbols in the standard library
            USED_SYMBOLS.add(symbol.name())

    global INCLUDES
    INCLUDES = includes

    grammars = []
    parsed_constraints: list[Constraint | SoftValue] = []
    if use_stdlib:
        assert STDLIB_GRAMMAR is not None
        assert STDLIB_CONSTRAINTS is not None
        try:
            grammars = [deepcopy(STDLIB_GRAMMAR)]
        except TypeError:
            # This can happen if we invoke parse() from a notebook
            grammars = [STDLIB_GRAMMAR]
        parsed_constraints = STDLIB_CONSTRAINTS.copy()

    grammars += given_grammars

    LOGGER.debug("Reading files")
    more_grammars = []
    global FILES_TO_PARSE
    for file in fan_files:
        if isinstance(file, str):
            FILES_TO_PARSE.append(("<string>", file, 0))  # TODO: fix
        else:
            FILES_TO_PARSE.append((file.name, file.read(), 0))

    global INCLUDE_DEPTH
    INCLUDE_DEPTH = 0

    mode = FuzzingMode.COMPLETE

    while FILES_TO_PARSE:
        (name, fan_contents, depth) = FILES_TO_PARSE.pop(0)
        LOGGER.debug(f"Reading {name} (depth = {depth})")
        new_grammar, new_constraints = parse_content(
            fan_contents,
            filename=name,
            use_cache=use_cache,
            lazy=lazy,
            max_repetitions=max_repetitions,
        )
        parsed_constraints += new_constraints
        assert new_grammar is not None
        if new_grammar.fuzzing_mode == FuzzingMode.IO:
            mode = FuzzingMode.IO

        if depth == 0:
            # Given file: process in order
            more_grammars.append(new_grammar)
            for generator in new_grammar.generators.values():
                for nonterminal in generator.nonterminals.values():
                    USED_SYMBOLS.add(nonterminal)
        else:
            # Included file: process _before_ current grammar
            more_grammars = [new_grammar] + more_grammars
            # Do not complain about unused symbols in included files
            for symbol in new_grammar.rules.keys():
                USED_SYMBOLS.add(symbol.name())
            for generator in new_grammar.generators.values():
                for nonterminal in generator.nonterminals.values():
                    USED_SYMBOLS.add(nonterminal.symbol.symbol)

        if INCLUDE_DEPTH > 0:
            INCLUDE_DEPTH -= 1

    grammars += more_grammars

    LOGGER.debug(f"Processing {len(grammars)} grammars")
    grammar = grammars[0]
    LOGGER.debug(f"Grammar #1: {[key.name() for key in grammar.rules.keys()]}")
    n = 2
    for g in grammars[1:]:
        LOGGER.debug(f"Grammar #{n}: {[key.name() for key in g.rules.keys()]}")

        for symbol in g.rules.keys():
            if symbol in grammar.rules:
                LOGGER.info(f"Redefining {symbol.name()}")

        grammar.update(g, prime=False)
        n += 1

    LOGGER.debug(f"Final grammar: {[key.name() for key in grammar.rules.keys()]}")

    grammar.fuzzing_mode = mode
    LOGGER.debug(f"Grammar fuzzing mode: {grammar.fuzzing_mode}")

    LOGGER.debug("Processing constraints")
    for constraint in constraints or []:
        LOGGER.debug(f"Constraint {constraint}")
        first_token = constraint.split()[0]
        if any(
            first_token.startswith(kw) for kw in ["where", "minimizing", "maximizing"]
        ):
            _, new_constraints = parse_content(
                constraint, filename=constraint, use_cache=use_cache, lazy=lazy
            )
        else:
            _, new_constraints = parse_content(
                "where " + constraint,
                filename=constraint,
                use_cache=use_cache,
                lazy=lazy,
            )
        parsed_constraints += new_constraints

    if check:
        LOGGER.debug("Checking and finalizing content")
        if grammar and len(grammar.rules) > 0:
            check_grammar_consistency(
                grammar, given_used_symbols=USED_SYMBOLS, start_symbol=start_symbol
            )

        if grammar and parsed_constraints:
            check_constraints_existence(grammar, parsed_constraints)

    global_env, local_env = grammar.get_spec_env()
    if not parties and grammar.fuzzing_mode == FuzzingMode.IO:
        # Prepare for interaction
        if "FandangoIO" not in global_env.keys():
            exec("FandangoIO.instance()", global_env, local_env)
        io_instance: FandangoIO = global_env["FandangoIO"].instance()

        assign_implicit_party(grammar, "StdOut")
        init_msg_parties(grammar, io_instance)
        remap_to_std_party(grammar, io_instance)
        init_msg_parties(grammar, io_instance)

        # Detect illegally nested data packets.
        rir_detector = MessageNestingDetector(grammar)
        rir_detector.fail_on_nested_packet(NonTerminal(start_symbol))
        fail_on_party_in_generator(grammar)

        truncate_invisible_packets(grammar, io_instance)

    # We invoke this at the very end, now that all data is there
    grammar.update(grammar, prime=check)

    if parties:
        slice_parties(grammar, parties)

    LOGGER.debug("All contents parsed")
    return grammar, parsed_constraints


### Consistency Checks


def fail_on_party_in_generator(grammar):
    for nt, node in grammar.rules.items():
        if nt not in grammar.generators:
            continue
        found_node = is_party_reachable(grammar, node)
        if found_node is not None:
            raise ValueError(
                f"{found_node} contains a party or recipient and is generated using the generator on {nt}. This is not allowed!"
            )

    for nt in grammar.generators.keys():
        dependencies: set[NonTerminal] = grammar.generator_dependencies(nt)
        for dep_nt in dependencies:
            found_node = is_party_reachable(grammar, grammar[dep_nt])
            if found_node is not None:
                raise ValueError(
                    f"{found_node} contains a party or recipient and is a parameter for the generator of {nt}. This is not allowed!"
                )


def is_party_reachable(grammar, node):
    seen_nt_nodes = set()
    symbol_finder = SymbolFinder()
    symbol_finder.visit(node)
    nt_node_queue: set[NonTerminalNode] = set(symbol_finder.nonTerminalNodes)
    while len(nt_node_queue) != 0:
        current_node = nt_node_queue.pop()
        if current_node.sender is not None or current_node.recipient is not None:
            return current_node

        seen_nt_nodes.add(current_node)
        symbol_finder = SymbolFinder()
        symbol_finder.visit(grammar[current_node.symbol])
        for next_nt in symbol_finder.nonTerminalNodes:
            if next_nt not in seen_nt_nodes:
                nt_node_queue.add(next_nt)
    return None


def init_msg_parties(
    grammar: "Grammar", io_instance: FandangoIO, ignore_existing: bool = True
):
    party_names = set()
    grammar_msg_parties = grammar.msg_parties(include_recipients=True)
    global_env, local_env = grammar.get_spec_env()

    # Initialize FandangoParty instances
    for key in global_env.keys():
        if key in grammar_msg_parties:
            the_type = global_env[key]
            if not isinstance(the_type, type):
                continue
            if FandangoParty in the_type.__mro__:
                party_names.add(key)
    # Call constructor
    for party in party_names:
        if party in io_instance.parties.keys() and ignore_existing:
            continue
        exec(f"{party}()", global_env, local_env)
        grammar_msg_parties.remove(party)


# Assign STD party to all parties which have no party-class defined.
def remap_to_std_party(grammar: "Grammar", io_instance: FandangoIO):
    remapped_parties = set()
    unknown_recipients = set()
    for symbol in grammar.rules.keys():
        symbol_finder = SymbolFinder()
        symbol_finder.visit(grammar.rules[symbol])
        non_terminals: list[NonTerminalNode] = symbol_finder.nonTerminalNodes

        for nt in non_terminals:
            if nt.sender is not None:
                if nt.sender not in io_instance.parties.keys():
                    remapped_parties.add(nt.sender)
                    nt.sender = "StdOut"
            if nt.recipient is not None:
                if nt.recipient not in io_instance.parties.keys():
                    unknown_recipients.add(nt.recipient)

    for name in remapped_parties:
        LOGGER.warning(f"Party {name!r} unspecified; will use 'StdOut' instead")
    if unknown_recipients:
        raise FandangoValueError(f"Recipients {unknown_recipients!r} unspecified")


def truncate_invisible_packets(grammar: "Grammar", io_instance: FandangoIO) -> None:
    keep_parties = grammar.msg_parties(include_recipients=True)
    io_instance.parties.keys()
    for existing_party in list(keep_parties):
        if not io_instance.parties[existing_party].is_fuzzer_controlled():
            keep_parties.remove(existing_party)

    for nt in grammar.rules.keys():
        PacketTruncator(grammar, keep_parties).visit(grammar.rules[nt])


def check_grammar_consistency(
    grammar, *, given_used_symbols=set(), start_symbol="<start>"
):
    check_grammar_definitions(
        grammar, given_used_symbols=given_used_symbols, start_symbol=start_symbol
    )
    check_grammar_types(grammar, start_symbol=start_symbol)


def check_grammar_definitions(
    grammar, *, given_used_symbols=set(), start_symbol="<start>"
):
    if not grammar:
        return

    LOGGER.debug("Checking grammar")

    used_symbols: set[str] = set()
    undefined_symbols: set[str] = set()
    defined_symbols: set[str] = set()

    for symbol in grammar.rules.keys():
        defined_symbols.add(symbol.name())

    if start_symbol not in defined_symbols:
        if start_symbol == "<start>":
            raise FandangoValueError(
                f"Start symbol {start_symbol!s} not defined in grammar"
            )
        closest = closest_match(start_symbol, defined_symbols)
        raise FandangoValueError(
            f"Start symbol {start_symbol!r} not defined in grammar. Did you mean {closest!r}?"
        )

    def collect_used_symbols(node: Node):
        if node.is_nonterminal:
            used_symbols.add(node.symbol.name())  # type: ignore[attr-defined] # We're checking types manually
        elif (
            node.node_type == NodeType.REPETITION
            or node.node_type == NodeType.STAR
            or node.node_type == NodeType.PLUS
            or node.node_type == NodeType.OPTION
        ):
            collect_used_symbols(node.node)  # type: ignore[attr-defined] # We're checking types manually

        for child in node.children():
            collect_used_symbols(child)

    for tree in grammar.rules.values():
        collect_used_symbols(tree)

    for symbol in used_symbols:
        if symbol not in defined_symbols:
            undefined_symbols.add(symbol)

    for symbol in defined_symbols:
        if (
            symbol not in used_symbols
            and symbol not in given_used_symbols
            and symbol != start_symbol
            and symbol != "<start>"  # Allow <start> to be defined but not used
        ):
            LOGGER.warning(f"Symbol {symbol!s} defined, but not used")

    if undefined_symbols:
        first_undefined_symbol = undefined_symbols.pop()
        error = FandangoValueError(
            f"Undefined symbol {first_undefined_symbol!s} in grammar"
        )
        if undefined_symbols:
            if getattr(Exception, "add_note", None):
                # Python 3.11+ has add_note() method
                error.add_note(
                    f"Other undefined symbols: {', '.join(undefined_symbols)}"
                )
        raise error


def check_grammar_types(
    grammar: Optional[Grammar], *, start_symbol: str = "<start>"
) -> None:
    if grammar is None:
        return

    LOGGER.debug("Checking types")

    symbol_types: dict[Symbol, tuple[Optional[str], int, int, int]] = {}

    def compatible(tp1, tp2):
        if tp1 in ["int", "bytes"] and tp2 in ["int", "bytes"]:
            return True
        return tp1 == tp2

    def get_type(tree: Node, rule_symbol: str) -> tuple[Optional[str], int, int, int]:
        # LOGGER.debug(f"Checking type of {tree!s} in {rule_symbol!s} ({tree.node_type!s})")
        nonlocal symbol_types, grammar

        tp: Optional[str]
        if isinstance(tree, TerminalNode):
            tp = type(tree.symbol).__name__
            # LOGGER.debug(f"Type of {tree.symbol.symbol!r} is {tp!r}")
            bits = 1 if tree.symbol.is_type(TreeValueType.TRAILING_BITS_ONLY) else 0
            return tp, bits, bits, 0

        elif (
            isinstance(tree, Repetition)
            or isinstance(tree, Star)
            or isinstance(tree, Plus)
            or isinstance(tree, Option)
        ):
            tp, min_bits, max_bits, step = get_type(tree.node, rule_symbol)
            # if min_bits % 8 != 0 and tree.min == 0:
            #     raise FandangoValueError(f"{rule_symbol!s}: Bits cannot be optional")

            rep_min = tree.min
            rep_max = tree.max

            step = min(min_bits, max_bits)
            return tp, rep_min * min_bits, rep_max * max_bits, step

        elif isinstance(tree, NonTerminalNode):
            if tree.symbol in symbol_types:
                return symbol_types[tree.symbol]

            symbol_types[tree.symbol] = (None, 0, 0, 0)
            assert grammar is not None
            symbol_tree = grammar.rules[tree.symbol]
            tp, min_bits, max_bits, step = get_type(symbol_tree, tree.symbol.name())
            symbol_types[tree.symbol] = tp, min_bits, max_bits, step
            # LOGGER.debug(f"Type of {tree.symbol!s} is {tp!r} with {min_bits}..{max_bits} bits")
            return tp, min_bits, max_bits, step

        elif (
            tree.node_type == NodeType.CONCATENATION
            or tree.node_type == NodeType.ALTERNATIVE
        ):
            common_tp = None
            tp_child = None
            first = True
            min_bits = 0
            max_bits = 0
            step = 0
            for child in tree.children():
                tp, min_child_bits, max_child_bits, child_step = get_type(
                    child, rule_symbol
                )
                if first:
                    min_bits = min_child_bits
                    max_bits = max_child_bits
                    step = child_step
                    first = False
                elif tree.node_type == NodeType.CONCATENATION:
                    min_bits += min_child_bits
                    max_bits += max_child_bits
                    step += child_step
                else:  # NodeType.ALTERNATIVE
                    min_bits = min(min_bits, min_child_bits)
                    max_bits = max(max_bits, max_child_bits)
                    step += min(step, child_step)
                if tp is None:
                    continue
                if common_tp is None:
                    common_tp = tp
                    tp_child = child
                    continue
                if not compatible(tp, common_tp):
                    if tree.node_type == NodeType.CONCATENATION:
                        LOGGER.warning(
                            f"{rule_symbol!s}: Concatenating {common_tp!r} ({tp_child!s}) and {tp!r} ({child!s})"
                        )
                    else:
                        LOGGER.warning(
                            f"{rule_symbol!s}: Type can be {common_tp!r} ({tp_child!s}) or {tp!r} ({child!s})"
                        )
                    common_tp = tp

            # LOGGER.debug(f"Type of {rule_symbol!s} is {common_tp!r} with {min_bits}..{max_bits} bits")
            return common_tp, min_bits, max_bits, step

        raise FandangoValueError("Unknown node type")

    start_tree = grammar.rules[NonTerminal(start_symbol)]
    _, min_start_bits, max_start_bits, start_step = get_type(start_tree, start_symbol)
    if start_step > 0 and any(
        bits % 8 != 0 for bits in range(min_start_bits, max_start_bits + 1, start_step)
    ):
        if min_start_bits != max_start_bits:
            LOGGER.warning(
                f"{start_symbol!s}: Number of bits ({min_start_bits}..{max_start_bits}) may not be a multiple of eight"
            )
        else:
            LOGGER.warning(
                f"{start_symbol!s}: Number of bits ({min_start_bits}) is not a multiple of eight"
            )


def check_constraints_existence(
    grammar: Grammar, constraints: list[Constraint | SoftValue]
):
    LOGGER.debug("Checking constraints")

    indirect_child: dict[str, dict[str, Optional[bool]]] = {
        k.name(): {l.name(): None for l in grammar.rules.keys()}  # noqa: E741
        for k in grammar.rules.keys()
    }

    defined_symbols = []
    for symbol in grammar.rules.keys():
        defined_symbols.append(symbol.name())

    grammar_symbols = grammar.rules.keys()
    grammar_matches = re.findall(
        r"<([^>]*)>", "".join(k.format_as_spec() for k in grammar_symbols)
    )
    # LOGGER.debug(f"All used symbols: {grammar_matches}")

    for constraint in constraints:
        constraint_symbols = constraint.get_symbols()

        for value in constraint_symbols:
            # LOGGER.debug(f"Constraint {constraint}: Checking {value}")

            constraint_matches = re.findall(
                r"<([^>]*)>", value.format_as_spec()
            )  # was <(.*?)>

            missing = [
                match for match in constraint_matches if match not in grammar_matches
            ]

            if missing:
                first_missing_symbol = missing[0]
                closest = closest_match(first_missing_symbol, defined_symbols)

            if len(missing) > 1:
                missing_symbols = ", ".join(
                    ["<" + str(symbol) + ">" for symbol in missing]
                )
                error = FandangoValueError(
                    f"{constraint}: undefined symbols {missing_symbols}. Did you mean {closest!r}?"
                )
                raise error

            if len(missing) == 1:
                missing_symbol = missing[0]
                error = FandangoValueError(
                    f"{constraint}: undefined symbol <{missing_symbol!r}>. Did you mean {closest!r}?"
                )
                raise error

            for i in range(len(constraint_matches) - 1):
                parent = constraint_matches[i]
                symbol = constraint_matches[i + 1]
                # This handles <parent>[...].<symbol> as <parent>..<symbol>.
                # We could also interpret the actual [...] contents here,
                # but slices and chains could make this hard -- AZ
                recurse = isinstance(value, DescendantAttributeSearch) or isinstance(
                    value, ItemSearch
                )
                if not check_constraints_existence_children(
                    grammar, parent, symbol, recurse, indirect_child
                ):
                    msg = f"{constraint!s}: <{parent!s}> has no child <{symbol!s}>"
                    raise FandangoValueError(msg)


def check_constraints_existence_children(
    grammar: Grammar,
    parent: str,
    symbol: str,
    recurse: bool,
    indirect_child: dict[str, dict[str, Optional[bool]]],
):
    # LOGGER.debug(f"Checking if <{symbol}> is a child of <{parent}>")

    if indirect_child[f"<{parent}>"][f"<{symbol}>"] is not None:
        return indirect_child[f"<{parent}>"][f"<{symbol}>"]

    grammar_symbols = grammar.rules[NonTerminal(f"<{parent}>")]

    # Original code; fails on <a> "b" <c> -- AZ
    # grammar_matches = re.findall(r'(?<!")<([^>]*)>(?!".*)',
    #                              str(grammar_symbols))
    #
    # Simpler version; may overfit (e.g. matches <...> in strings),
    # but that should not hurt us -- AZ
    finder = SymbolFinder()
    finder.visit(grammar_symbols)
    non_terminals = [nt.symbol.name()[1:-1] for nt in finder.nonTerminalNodes]

    if symbol in non_terminals:
        indirect_child[f"<{parent}>"][f"<{symbol}>"] = True
        return True

    is_child = False
    for match in non_terminals:
        if recurse or match.startswith("_"):
            is_child = is_child or check_constraints_existence_children(
                grammar, match, symbol, recurse, indirect_child
            )
    indirect_child[f"<{parent}>"][f"<{symbol}>"] = is_child
    return is_child


def slice_parties(grammar: "Grammar", parties: list[str]) -> None:
    """
    Slice the given parties from the grammar.
    :param grammar: The grammar to check
    :param parties: List of party names to check
    :raises FandangoValueError: If a party is not defined in the grammar
    """
    if not parties:
        return

    defined_parties = set(grammar.msg_parties(include_recipients=True))
    for party in parties:
        if party not in defined_parties:
            closest = closest_match(party, defined_parties)
            raise FandangoValueError(
                f"Party {party!r} not defined in the grammar. Did you mean {closest!r}?"
            )

    grammar.slice_parties(parties)


def assign_implicit_party(grammar, implicit_party: str):
    seen_nts: set[NonTerminal] = set()
    seen_nts.add(NonTerminal("<start>"))
    processed_nts: set[NonTerminal] = set()
    unprocessed_nts: set[NonTerminal] = seen_nts.difference(processed_nts)

    while len(unprocessed_nts) > 0:
        current_symbol = unprocessed_nts.pop()
        current_node = grammar.rules[current_symbol]

        symbol_finder = SymbolFinder()
        symbol_finder.visit(current_node)
        rule_nts = list(
            filter(lambda x: x not in processed_nts, symbol_finder.nonTerminalNodes)
        )

        if current_node in rule_nts and not isinstance(current_node, NonTerminalNode):
            rule_nts.remove(current_node)
        child_party: set[str] = set()

        for c_node in rule_nts:
            child_party |= c_node.msg_parties(include_recipients=False)

        if len(child_party) == 0:
            processed_nts.add(current_symbol)
            unprocessed_nts = seen_nts.difference(processed_nts)
            continue
        for c_node in rule_nts:
            seen_nts.add(c_node.symbol)
            if len(c_node.msg_parties(include_recipients=False)) != 0:
                continue
            c_node.sender = implicit_party
        for t_node in symbol_finder.terminalNodes:
            terminal_id = 0
            rule_nt = NonTerminal(f"<_terminal:{terminal_id}>")
            while rule_nt in grammar.rules:
                terminal_id += 1
                rule_nt = NonTerminal(f"<_terminal:{terminal_id}>")
            n_node = NonTerminalNode(
                rule_nt,
                grammar.grammar_settings,
                implicit_party,
            )
            NodeReplacer(t_node, n_node).visit(current_node)
            grammar.rules[rule_nt] = t_node

        processed_nts.add(current_symbol)
        unprocessed_nts = seen_nts.difference(processed_nts)



from typing import Optional
from fandango.language.grammar.node_visitors.node_visitor import NodeVisitor
from fandango.language.grammar.nodes.alternative import Alternative
from fandango.language.grammar.nodes.concatenation import Concatenation
from fandango.language.grammar.nodes.node import Node
from fandango.language.grammar.nodes.non_terminal import NonTerminalNode
from fandango.language.grammar.nodes.repetition import Repetition, Star, Plus, Option
from fandango.language.grammar.nodes.terminal import TerminalNode


class NodeReplacer(NodeVisitor):
    def __init__(self, old_node: Node, new_node: Node):
        self.old_node = old_node
        self.new_node = new_node

    def replace(self, node: Node) -> Node:
        if node == self.old_node:
            return self.new_node
        return node

    def default_result(self) -> list[Node]:
        return []

    def aggregate_results(
        self, aggregate: list[Node], result: Optional[Node]
    ) -> list[Node]:
        if result is not None:
            aggregate.append(result)
        return aggregate

    def visitConcatenation(self, node: Concatenation) -> Node:
        replaced = self.replace(node)
        if isinstance(replaced, Concatenation):
            replaced.nodes = self.visitChildren(replaced)
            return replaced
        else:
            return self.visit(replaced)

    def visitAlternative(self, node: Alternative) -> Node:
        replaced = self.replace(node)
        if isinstance(replaced, Alternative):
            replaced.alternatives = self.visitChildren(replaced)
            return replaced
        else:
            return self.visit(replaced)

    def visitRepetition(self, node: Repetition) -> Node:
        replaced = self.replace(node)
        if isinstance(replaced, Repetition):
            replaced.node = self.visit(replaced.node)
            return replaced
        else:
            return self.visit(replaced)

    def visitStar(self, node: Star) -> Node:
        replaced = self.replace(node)
        if isinstance(replaced, Star):
            replaced.node = self.visit(replaced.node)
            return replaced
        else:
            return self.visit(replaced)

    def visitPlus(self, node: Plus) -> Node:
        replaced = self.replace(node)
        if isinstance(replaced, Plus):
            replaced.node = self.visit(replaced.node)
            return replaced
        else:
            return self.visit(replaced)

    def visitOption(self, node: Option) -> Node:
        replaced = self.replace(node)
        if isinstance(replaced, Option):
            replaced.node = self.visit(replaced.node)
            return replaced
        else:
            return self.visit(replaced)

    def visitNonTerminalNode(self, node: NonTerminalNode) -> Node:
        replaced = self.replace(node)
        if isinstance(replaced, NonTerminalNode):
            return replaced
        else:
            return self.visit(replaced)

    def visitTerminalNode(self, node: TerminalNode) -> Node:
        replaced = self.replace(node)
        if isinstance(replaced, TerminalNode):
            return replaced
        else:
            return self.visit(replaced)



from typing import Optional
from fandango.language.tree import DerivationTree
from fandango.language.symbols.symbol import Symbol


class ParserDerivationTree(DerivationTree):
    def __init__(
        self,
        symbol: Symbol,
        children: Optional[list[DerivationTree]] = None,
        *,
        parent: Optional[DerivationTree] = None,
        sender: Optional[str] = None,
        recipient=None,
        read_only: bool = False,
        origin_repetitions: list[tuple[str, int, int]] | None = None,
    ):
        super().__init__(
            symbol,
            children,
            parent=parent,
            sources=[],
            sender=sender,
            recipient=recipient,
            read_only=read_only,
            origin_repetitions=origin_repetitions,
        )

    def set_children(self, children: list[DerivationTree]):
        self._children = children
        self.invalidate_hash()



import random
from typing import Counter, Union
from collections.abc import Generator

from fandango.constraints.constraint import Constraint
from fandango.constraints.soft import SoftValue
from fandango.constraints.fitness import FailingTree
from fandango.language import DerivationTree, Grammar
from fandango.logger import LOGGER, print_exception


class Evaluator:
    def __init__(
        self,
        grammar: Grammar,
        constraints: list[Union[Constraint, SoftValue]],
        expected_fitness: float,
        diversity_k: int,
        diversity_weight: float,
        warnings_are_errors: bool = False,
    ):
        self._grammar = grammar
        self._soft_constraints: list[SoftValue] = []
        self._hard_constraints: list[Constraint] = []
        self._expected_fitness = expected_fitness
        self._diversity_k = diversity_k
        self._diversity_weight = diversity_weight
        self._warnings_are_errors = warnings_are_errors
        self._fitness_cache: dict[int, tuple[float, list[FailingTree]]] = {}
        self._solution_set: set[int] = set()
        self._checks_made = 0

        for constraint in constraints:
            if isinstance(constraint, SoftValue):
                self._soft_constraints.append(constraint)
            elif isinstance(constraint, Constraint):
                self._hard_constraints.append(constraint)
            else:
                raise ValueError(f"Invalid constraint type: {type(constraint)}")

    @property
    def expected_fitness(self) -> float:
        return self._expected_fitness

    def get_fitness_check_count(self) -> int:
        """
        :return: The number of fitness checks made so far.
        """
        return self._checks_made

    def compute_mutation_pool(
        self, population: list[DerivationTree]
    ) -> list[DerivationTree]:
        """
        Computes the mutation pool for the given population.

        The mutation pool is computed by sampling the population with replacement, where the probability of sampling an individual is proportional to its fitness.

        :param population: The population to compute the mutation pool for.
        :return: The mutation pool.
        """
        weights = [self._fitness_cache[hash(ind)][0] for ind in population]
        if not all(w == 0 for w in weights):
            return random.choices(population, weights=weights, k=len(population))
        else:
            return population

    def flush_fitness_cache(self) -> None:
        """
        For soft constraints, the normalized fitness may change over time as we observe more inputs, this method flushes the fitness cache if the grammar contains any soft constraints.
        """
        if len(self._soft_constraints) > 0:
            self._fitness_cache = {}

    def compute_diversity_bonus(self, individuals: list[DerivationTree]) -> list[float]:
        ind_kpaths = [
            self._grammar._extract_k_paths_from_tree(ind, self._diversity_k)
            for ind in individuals
        ]
        frequencies = Counter(path for paths in ind_kpaths for path in paths)

        bonus = [
            (
                sum(1.0 / frequencies[path] for path in paths) / len(paths)
                if paths
                else 0.0
            )
            for paths in ind_kpaths
        ]
        return bonus

    def evaluate_hard_constraints(
        self, individual: DerivationTree
    ) -> tuple[float, list[FailingTree]]:
        if len(self._hard_constraints) == 0:
            return 1.0, []

        hard_fitness = 0.0
        failing_trees: list[FailingTree] = []
        for constraint in self._hard_constraints:
            try:
                result = constraint.fitness(individual)

                if result.success:
                    hard_fitness += result.fitness()
                else:
                    failing_trees.extend(result.failing_trees)
                    hard_fitness += result.fitness()
                self._checks_made += 1
            except Exception as e:
                LOGGER.error(
                    f"Error evaluating hard constraint {constraint.format_as_spec()}"
                )
                print_exception(e)
                hard_fitness += 0.0
        hard_fitness /= len(self._hard_constraints)
        return hard_fitness, failing_trees

    def evaluate_soft_constraints(
        self, individual: DerivationTree
    ) -> tuple[float, list[FailingTree]]:
        soft_fitness = 0.0
        failing_trees: list[FailingTree] = []
        for constraint in self._soft_constraints:
            try:
                result = constraint.fitness(individual)

                # failing_trees are required for mutations;
                # with soft constraints, we never know when they are fully optimized.
                failing_trees.extend(result.failing_trees)

                constraint.tdigest.update(result.fitness())
                normalized_fitness = constraint.tdigest.score(result.fitness())

                if constraint.optimization_goal == "max":
                    soft_fitness += normalized_fitness
                else:  # "min"
                    soft_fitness += 1 - normalized_fitness
            except Exception as e:
                LOGGER.error(f"Error evaluating soft constraint {constraint}: {e}")
                print_exception(e)
                soft_fitness += 0.0

        soft_fitness /= len(self._soft_constraints)
        return soft_fitness, failing_trees

    def evaluate_individual(
        self,
        individual: DerivationTree,
    ) -> Generator[DerivationTree, None, tuple[float, list[FailingTree]]]:
        key = hash(individual)
        if key in self._fitness_cache:
            return self._fitness_cache[key]

        fitness, failing_trees = self.evaluate_hard_constraints(individual)

        if self._soft_constraints:
            if fitness < 1.0:
                fitness = (
                    fitness
                    * len(self._hard_constraints)
                    / (len(self._hard_constraints) + len(self._soft_constraints))
                )
            else:  # fitness from hard constraints == 1.0
                soft_fitness, soft_failing_trees = self.evaluate_soft_constraints(
                    individual
                )

                failing_trees.extend(soft_failing_trees)

                fitness = (
                    fitness * len(self._hard_constraints)
                    + soft_fitness * len(self._soft_constraints)
                ) / (len(self._hard_constraints) + len(self._soft_constraints))

        if fitness >= self._expected_fitness and key not in self._solution_set:
            self._solution_set.add(key)
            yield individual

        self._fitness_cache[key] = (fitness, failing_trees)
        return fitness, failing_trees

    def evaluate_population(
        self,
        population: list[DerivationTree],
    ) -> Generator[
        DerivationTree, None, list[tuple[DerivationTree, float, list[FailingTree]]]
    ]:
        evaluation: list[tuple[DerivationTree, float, list[FailingTree]]] = []
        for ind in population:
            ind_eval = yield from self.evaluate_individual(ind)
            evaluation.append((ind, *ind_eval))

        if self._diversity_k > 0 and self._diversity_weight > 0:
            bonuses = self.compute_diversity_bonus(population)
            evaluation = [
                (ind, fitness + bonus, failing_trees)
                for (ind, fitness, failing_trees), bonus in zip(evaluation, bonuses)
            ]

        return evaluation

    def select_elites(
        self,
        evaluation: list[tuple[DerivationTree, float, list[FailingTree]]],
        elitism_rate: float,
        population_size: int,
    ) -> list[DerivationTree]:
        return [
            x[0]
            for x in sorted(evaluation, key=lambda x: x[1], reverse=True)[
                : int(elitism_rate * population_size)
            ]
        ]

    def tournament_selection(
        self,
        evaluation: list[tuple[DerivationTree, float, list[FailingTree]]],
        tournament_size: int,
    ) -> tuple[DerivationTree, DerivationTree]:
        tournament = random.sample(evaluation, k=min(tournament_size, len(evaluation)))
        tournament.sort(key=lambda x: x[1], reverse=True)
        parent1 = tournament[0][0]
        if len(tournament) == 2:
            parent2 = tournament[1][0] if tournament[1][0] != parent1 else parent1
        else:
            parent2 = (
                tournament[1][0] if tournament[1][0] != parent1 else tournament[2][0]
            )
        return parent1, parent2



#!/usr/bin/env pytest

from collections.abc import Generator
import itertools
import unittest

from fandango.evolution.algorithm import Fandango, LoggerLevel
from fandango.language.parse import parse
from .utils import RESOURCES_ROOT, run_command


class TestSoft(unittest.TestCase):
    @staticmethod
    def get_solutions(
        specification_file, desired_solutions, random_seed, max_generations=500
    ) -> Generator[str, None, None]:
        with open(specification_file, "r") as file:
            grammar_int, constraints_int = parse(
                file, use_stdlib=False, use_cache=False
            )
        assert grammar_int is not None
        assert constraints_int is not None
        fandango = Fandango(
            grammar=grammar_int,
            constraints=constraints_int,
            random_seed=random_seed,
            logger_level=LoggerLevel.DEBUG,
        )
        generator = itertools.islice(
            fandango.generate(max_generations=max_generations), desired_solutions
        )

        for solution in generator:
            yield str(solution)


class TestSoftValue(TestSoft):
    def test_soft_value(self):
        gen = self.get_solutions(
            RESOURCES_ROOT / "softvalue.fan",
            desired_solutions=100,
            random_seed=1,
        )
        solutions = []
        for s in gen:
            solutions.append(s)
            if "999999-999999" == s:
                return  # optimized solution found, stop generating and don't fail

        self.fail(
            f"999999-999999 not found in the first {len(solutions)} solutions, found: {solutions}"
        )

    def test_min_in_different_contexts(self):
        gen = self.get_solutions(
            RESOURCES_ROOT / "persons_with_constr.fan",
            desired_solutions=60,
            random_seed=1,
        )
        for solution in gen:
            name, age = solution.split(",")
            first_name, last_name = name.split(" ")
            if len(first_name) == 2 and len(last_name) == 2:
                return
        self.fail(
            "No solution found, last first_name: {first_name}, last last_name: {last_name}"
        )

    def test_cli_max_1(self):
        command = [
            "fandango",
            "fuzz",
            "-f",
            str(RESOURCES_ROOT / "persons.fan"),
            "-c",
            "maximizing int(<age>)",
            "-n",
            "50",
            "--random-seed",
            "1",
        ]
        out, err, code = run_command(command)
        lines = [line for line in out.split("\n") if line.strip()]
        self.assertGreater(len(lines), 0, f"\nerr: {err}\nout: {out}")
        last_age = int(lines[-1].split(",")[1])  # e.g., 9999999999999599999999
        self.assertGreater(last_age, 9999999999999, f"\nerr: {err}\nout: {out}")

    def test_cli_max_2(self):
        command = [
            "fandango",
            "fuzz",
            "-f",
            str(RESOURCES_ROOT / "persons.fan"),
            "--maximize",
            "int(<age>)",
            "-n",
            "50",
            "--random-seed",
            "1",
        ]
        out, err, code = run_command(command)
        lines = [line for line in out.split("\n") if line.strip()]
        last_age = int(lines[-1].split(",")[1])  # e.g., 9999999999999599999999
        self.assertGreater(last_age, 9999999999999)

    def test_cli_min_1(self):
        command = [
            "fandango",
            "fuzz",
            "-f",
            str(RESOURCES_ROOT / "persons.fan"),
            "-c",
            "minimizing int(<age>)",
            "-n",
            "100",
            "--random-seed",
            "1",
        ]
        out, err, code = run_command(command)
        lines = [line for line in out.split("\n") if line.strip()]
        last_age = int(lines[-1].split(",")[1])
        self.assertEqual(last_age, 0)

    def test_cli_min_2(self):
        command = [
            "fandango",
            "fuzz",
            "-f",
            str(RESOURCES_ROOT / "persons.fan"),
            "--minimize",
            "int(<age>)",
            "-n",
            "100",
            "--random-seed",
            "1",
        ]
        out, err, code = run_command(command)
        lines = [line for line in out.split("\n") if line.strip()]
        last_age = int(lines[-1].split(",")[1])
        self.assertEqual(last_age, 0)



import argparse
import ctypes
from io import UnsupportedOperation
import os
import subprocess
import sys
import tempfile
from typing import IO, Any
from fandango.language.tree import DerivationTree
from fandango.logger import LOGGER, clear_visualization


def output(
    tree: DerivationTree, args: argparse.Namespace, file_mode: str
) -> str | bytes:
    assert file_mode == "binary" or file_mode == "text"

    if args.format == "string":
        if file_mode == "binary":
            LOGGER.debug("Output as bytes")
            return tree.to_bytes()
        elif file_mode == "text":
            LOGGER.debug("Output as text")
            return tree.to_string()

    def convert(s: str) -> str | bytes:
        if file_mode == "binary":
            return s.encode("utf-8")
        else:
            return s

    LOGGER.debug(f"Output as {args.format}")

    if args.format == "tree":
        return convert(tree.to_tree())
    if args.format == "repr":
        return convert(tree.to_repr())
    if args.format == "bits":
        return convert(tree.to_bits())
    if args.format == "grammar":
        return convert(tree.to_grammar())
    if args.format == "value":
        return convert(tree.to_value())
    if args.format == "none":
        return convert("")

    raise NotImplementedError("Unsupported output format")


def open_file(filename: str, file_mode: str, *, mode: str = "r") -> IO[Any]:
    assert file_mode == "binary" or file_mode == "text"

    if file_mode == "binary":
        mode += "b"

    LOGGER.debug(f"Opening {filename!r}; mode={mode!r}")

    if filename == "-":
        if "b" in mode:
            return sys.stdin.buffer if "r" in mode else sys.stdout.buffer
        else:
            return sys.stdin if "r" in mode else sys.stdout

    return open(filename, mode)


def output_population(
    population: list[DerivationTree],
    args: argparse.Namespace,
    file_mode: str,
    *,
    output_on_stdout: bool = True,
) -> None:
    if args.format == "none":
        return

    for i, solution in enumerate(population):
        output_solution(solution, args, i, file_mode, output_on_stdout=output_on_stdout)


def output_solution_to_directory(
    solution: DerivationTree,
    args: argparse.Namespace,
    solution_index: int,
    file_mode: str,
) -> None:
    LOGGER.debug(f"Storing solution in directory {args.directory!r}")
    os.makedirs(args.directory, exist_ok=True)

    basename = f"fandango-{solution_index:04d}{args.filename_extension}"
    filename = os.path.join(args.directory, basename)
    with open_file(filename, file_mode, mode="w") as fd:
        fd.write(output(solution, args, file_mode))


def output_solution_to_file(
    solution: DerivationTree,
    args: argparse.Namespace,
    file_mode: str,
) -> None:
    LOGGER.debug(f"Storing solution in file {args.output!r}")
    with open_file(args.output, file_mode, mode="a") as fd:
        try:
            position = fd.tell()
        except (UnsupportedOperation, OSError):
            # If we're writing to stdout, tell() may not be supported
            position = 0

        if position > 0:
            fd.write(
                args.separator.encode("utf-8")
                if file_mode == "binary"
                else args.separator
            )
        fd.write(output(solution, args, file_mode))


def output_solution_with_test_command(
    solution: DerivationTree, args: argparse.Namespace, file_mode: str
) -> None:
    LOGGER.info(f"Running {args.test_command}")
    base_cmd = [args.test_command] + args.test_args

    if args.input_method == "filename":
        prefix = "fandango-"
        suffix = args.filename_extension
        mode = "wb" if file_mode == "binary" else "w"

        # The return type is private, so we need to use Any
        def named_temp_file(*, mode: str, prefix: str, suffix: str) -> Any:
            try:
                # Windows needs delete_on_close=False, so the subprocess can access the file by name
                return tempfile.NamedTemporaryFile(  # type: ignore [call-overload] # the mode type is not available from the library
                    mode=mode,
                    prefix=prefix,
                    suffix=suffix,
                    delete_on_close=False,
                )
            except Exception:
                # Python 3.11 and earlier have no 'delete_on_close'
                return tempfile.NamedTemporaryFile(
                    mode=mode, prefix=prefix, suffix=suffix
                )

        with named_temp_file(mode=mode, prefix=prefix, suffix=suffix) as fd:
            fd.write(output(solution, args, file_mode))
            fd.flush()
            cmd = base_cmd + [fd.name]
            LOGGER.debug(f"Running {cmd}")
            subprocess.run(cmd, text=True)
    elif args.input_method == "stdin":
        cmd = base_cmd
        LOGGER.debug(f"Running {cmd} with individual as stdin")
        subprocess.run(
            cmd,
            input=output(solution, args, file_mode),
            text=(None if file_mode == "binary" else True),
        )
    elif args.input_method == "libfuzzer":
        if args.file_mode != "binary" or file_mode != "binary":
            raise NotImplementedError("LibFuzzer harnesses only support binary input")
        harness = ctypes.CDLL(args.test_command).LLVMFuzzerTestOneInput

        bytes = output(solution, args, file_mode)
        harness(bytes, len(bytes))
    else:
        raise NotImplementedError("Unsupported input method")


def output_solution_to_stdout(
    solution: DerivationTree,
    args: argparse.Namespace,
    file_mode: str,
) -> None:
    LOGGER.debug("Printing solution on stdout")
    out = output(solution, args, file_mode)
    if not isinstance(out, str):
        out = out.decode("iso8859-1")
    print(out, end="")
    print(args.separator, end="")


def output_solution(
    solution: DerivationTree,
    args: argparse.Namespace,
    solution_index: int,
    file_mode: str,
    *,
    output_on_stdout: bool = True,
) -> None:
    assert file_mode == "binary" or file_mode == "text"

    if args.format == "none":
        return
    if "output" not in args:
        return

    if args.directory:
        output_solution_to_directory(solution, args, solution_index, file_mode)
        output_on_stdout = False

    if args.output:
        output_solution_to_file(solution, args, file_mode)
        output_on_stdout = False

    if "test_command" in args and args.test_command:
        output_solution_with_test_command(solution, args, file_mode)
        output_on_stdout = False

    # Default
    if output_on_stdout:
        clear_visualization()
        output_solution_to_stdout(solution, args, file_mode)



class FandangoError(ValueError):
    """Generic Error"""

    pass


class FandangoParseError(FandangoError, SyntaxError):
    """Error during parsing inputs"""

    def __init__(self, message: str | None = None, position: int | None = None):
        if message is None:
            if position is not None:
                message = f"Parse error at position {position}"
            else:
                message = "Parse error"

        # Call the parent class constructors
        FandangoError.__init__(self, message)
        SyntaxError.__init__(self, message)
        self.position = position


class FandangoSyntaxError(FandangoError, SyntaxError):
    """Error during parsing a Fandango spec"""

    def __init__(self, message: str):
        FandangoError.__init__(self, message)
        SyntaxError.__init__(self, message)


class FandangoValueError(FandangoError, ValueError):
    """Error during evaluating a Fandango spec"""

    def __init__(self, message: str):
        FandangoError.__init__(self, message)
        ValueError.__init__(self, message)


class FandangoConversionError(FandangoValueError):
    """
    An error raised when a conversion call on a TreeValue fails.
    """

    def __init__(self, message: str):
        super().__init__(message)


class FandangoFailedError(FandangoError):
    """Error during the Fandango algorithm"""

    def __init__(self, message: str):
        super().__init__(self, message)



import time
from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import Any, Union
from collections.abc import Generator


class Timer(ABC):
    """Abstract base class for timing operations."""

    def __init__(self, profiler: "Profiler", key: str):
        self.profiler = profiler
        self.key = key
        self._start_time = time.time()

    @abstractmethod
    def increment(self, count: int = 1) -> None:
        """Increment the count for this metric."""
        pass

    @abstractmethod
    def _stop(self) -> None:
        """Stop the timer and record the elapsed time."""
        pass


class EnabledTimer(Timer):
    """Timer implementation for when profiling is enabled."""

    def _stop(self) -> None:
        elapsed = time.time() - self._start_time
        if "time" not in self.profiler.metrics[self.key]:
            self.profiler.metrics[self.key]["time"] = 0
        self.profiler.metrics[self.key]["time"] += elapsed

    def increment(self, count: int = 1) -> None:
        if not self.profiler.enabled:
            return

        if "count" not in self.profiler.metrics[self.key]:
            self.profiler.metrics[self.key]["count"] = 0
        self.profiler.metrics[self.key]["count"] += count


class DisabledTimer(Timer):
    """No-op timer implementation for when profiling is disabled."""

    def __init__(self, profiler: "Profiler", key: str):
        pass

    def increment(self, count: int = 1) -> None:
        pass

    def _stop(self) -> None:
        pass


class Profiler:
    """A profiling utility for tracking execution times and counts."""

    def __init__(self, enabled: bool):
        self.enabled = enabled
        self.metrics: dict[str, dict[str, Union[int, float]]] = {
            "initial_population": {"count": 0, "time": 0.0},
            "evaluate_population": {"count": 0, "time": 0.0},
            "select_elites": {"count": 0, "time": 0.0},
            "tournament_selection": {"count": 0, "time": 0.0},
            "filling": {"count": 0, "time": 0.0},
            "crossover": {"count": 0, "time": 0.0},
            "mutation": {"count": 0, "time": 0.0},
        }

    @contextmanager
    def timer(
        self, key: str, increment: Union[int, list[Any], None] = None
    ) -> Generator[Timer, None, None]:
        """Context manager for profiling operations.

        :param key: The metric key to track
        :param increment: Either an integer, a list the length which will be used as the increment value.
        :yields: A timer object that can be used to increment the metric if it has to be calculated manually.
        """
        timer = EnabledTimer(self, key) if self.enabled else DisabledTimer(self, key)
        try:
            yield timer
        finally:
            if self.enabled:
                timer._stop()
                # Calculate increment value after operation completes
                if isinstance(increment, list):
                    timer.increment(len(increment))
                elif isinstance(increment, int):
                    timer.increment(increment)
                elif increment is not None:
                    raise ValueError(f"Invalid increment value: {increment}")

    def log_results(self) -> None:
        """Log the profiling results."""
        if not self.enabled:
            return

        for key, value in self.metrics.items():
            if isinstance(value, dict) and "time" in value and "count" in value:
                avg_time = value["time"] / value["count"] if value["count"] > 0 else 0
                print(
                    f"{key}: {avg_time:.6f}s per execution ({value['count']} runs, total {value['time']:.6f}s)"
                )
            else:
                print(f"Warning: '{key}' does not have valid time/count data.")



#!/usr/bin/env pytest

import unittest

from .utils import DOCS_ROOT, PROJECT_ROOT, run_command


class test_convert(unittest.TestCase):
    def test_convert_fan(self):
        command = ["fandango", "convert", str(DOCS_ROOT / "persons.fan")]
        out, err, code = run_command(command)
        self.assertEqual(0, code, f"Command failed with code {code}: {err}")
        self.assertEqual(err, "")

    def test_convert_antlr(self):
        calculator = (
            PROJECT_ROOT / "src" / "fandango" / "converters" / "antlr" / "Calculator.g4"
        )
        command = [
            "fandango",
            "convert",
            str(calculator),
        ]
        out, err, code = run_command(command)
        self.assertEqual(0, code)
        self.assertEqual(err, "")

    def test_convert_dtd(self):
        svg = (
            PROJECT_ROOT
            / "src"
            / "fandango"
            / "converters"
            / "dtd"
            / "svg11-flat-20110816.dtd"
        )
        command = [
            "fandango",
            "convert",
            str(svg),
        ]
        out, err, code = run_command(command)
        self.assertEqual(0, code, f"Command failed with code {code}: {err}")
        self.assertEqual(err, "")

    def test_convert_bt(self):
        gif = PROJECT_ROOT / "src" / "fandango" / "converters" / "bt" / "gif.bt"
        command = [
            "fandango",
            "convert",
            "--endianness=little",
            "--bitfield-order=left-to-right",
            str(gif),
        ]
        out, err, code = run_command(command)
        self.assertEqual(0, code)
        self.assertEqual(err, "")

    def test_convert_bt_again(self):
        gif = PROJECT_ROOT / "src" / "fandango" / "converters" / "bt" / "gif.bt"
        command = [
            "fandango",
            "convert",
            "--endianness=big",
            "--bitfield-order=right-to-left",
            str(gif),
        ]
        out, err, code = run_command(command)
        self.assertEqual(0, code)
        self.assertEqual(err, "")



#!/usr/bin/env pytest

from collections.abc import Generator
from copy import deepcopy
import itertools
import random
import unittest

from fandango.constraints.fitness import FailingTree
from fandango.evolution import GeneratorWithReturn
from fandango.evolution.algorithm import Fandango, LoggerLevel
from fandango.evolution.population import PopulationManager
from fandango.language.parse import parse
from fandango.language.tree import DerivationTree
from .utils import RESOURCES_ROOT


class GeneticTest(unittest.TestCase):
    def is_this_art_or_can_it_be_deleted(self):
        # I just need to keep this part of the code to ensure such a piece of art is not lost in the ether
        #
        # From a fellow connoisseur of art: thank you for preserving this!
        #
        # Define a simple grammar for testing
        file = open("tests/resources/example_number.fan", "r")
        try:
            grammar_int, constraints_int = parse(
                file, use_stdlib=False, use_cache=False
            )
        except FileNotFoundError:
            grammar_int, constraints_int = parse(
                file, use_stdlib=False, use_cache=False
            )

        random.seed(25)  # Set random seed

        # Initialize FANDANGO with a fixed random seed for reproducibility
        assert grammar_int is not None
        self.fandango = Fandango(
            grammar=grammar_int,
            constraints=constraints_int,
            population_size=50,
            mutation_rate=0.2,
            crossover_rate=0.8,
            elitism_rate=0.2,
            logger_level=LoggerLevel.DEBUG,
        )

    def setUp(self):
        with open(RESOURCES_ROOT / "example_number.fan", "r") as f:
            grammar_int, constraints_int = parse(f, use_stdlib=False, use_cache=False)
        assert grammar_int is not None
        random.seed(25)  # Set random seed for reproducibility
        self.fandango = Fandango(
            grammar=grammar_int,
            constraints=constraints_int,
            population_size=50,
            mutation_rate=0.2,
            crossover_rate=0.8,
            elitism_rate=0.2,
            logger_level=LoggerLevel.DEBUG,
        )
        list(
            self.fandango.generate_initial_population()
        )  # ensure the generator runs until the end

    def test_refill_population_with_empty_population(self):
        manager = PopulationManager(
            grammar=self.fandango.grammar,
            start_symbol=self.fandango.start_symbol,
            warnings_are_errors=True,
        )
        population: list[DerivationTree] = []
        expected_count = 10
        generator = manager.refill_population(
            current_population=population,
            eval_individual=self.fandango.evaluator.evaluate_individual,
            max_nodes=self.fandango.current_max_nodes,
            target_population_size=expected_count,
        )
        solutions = list(generator)
        self.assertLessEqual(len(solutions), expected_count)
        self.assertEqual(len(population), expected_count)
        for individual in population:
            self.assertIsInstance(individual, DerivationTree)
            self.assertTrue(self.fandango.grammar.parse(str(individual)))
        for individual in solutions:
            self.assertIn(individual, population)
            self.assertIsInstance(individual, DerivationTree)
            self.assertTrue(self.fandango.grammar.parse(str(individual)))

    def test_refill_population_with_non_empty_population(self):
        # Generate a population of derivation trees
        manager = PopulationManager(
            grammar=self.fandango.grammar,
            start_symbol=self.fandango.start_symbol,
            warnings_are_errors=True,
        )

        initial_count = 10
        population: list[DerivationTree] = []

        # add some initial individuals
        generator = manager.refill_population(
            current_population=population,
            eval_individual=self.fandango.evaluator.evaluate_individual,
            max_nodes=self.fandango.current_max_nodes,
            target_population_size=initial_count,
        )

        _initial_solutions = list(generator)  # drain initial solutions

        copy_of_initial_population = deepcopy(population)

        additional_count = 10
        generator = manager.refill_population(
            current_population=population,
            eval_individual=self.fandango.evaluator.evaluate_individual,
            max_nodes=self.fandango.current_max_nodes,
            target_population_size=initial_count + additional_count,
        )
        solutions = list(generator)

        self.assertEqual(len(population), initial_count + additional_count)
        self.assertLessEqual(len(solutions), additional_count)
        for individual in solutions:
            self.assertIn(individual, population)
            self.assertIsInstance(individual, DerivationTree)
            self.assertTrue(self.fandango.grammar.parse(str(individual)))
        for individual in copy_of_initial_population:
            self.assertIn(individual, population)

    def test_evaluate_fitness(self):
        # Evaluate the fitness of the population
        population = self.fandango.population
        unique = set(population)
        for individual in population:
            generator = GeneratorWithReturn(
                self.fandango.evaluator.evaluate_individual(individual)
            )
            solutions = list(generator)  # extract all solutions
            if individual not in unique:
                individual_possibly_int = individual.to_int()
                assert individual_possibly_int is not None
                individual_int = individual_possibly_int % 2
            else:
                individual_int = 0
            self.assertEqual(
                len(solutions), individual_int, f"Individual: {individual}"
            )
            fitness, failing_trees = generator.return_value
            self.assertIsInstance(fitness, float)
            self.assertGreaterEqual(fitness, 0.0)
            self.assertLessEqual(fitness, 1.0)
            self.assertIsInstance(failing_trees, list)
            for failing_tree in failing_trees:
                self.assertIsInstance(failing_tree, FailingTree)
            self.assertTrue(self.fandango.grammar.parse(str(individual)))

    def test_evaluate_population(self):
        # Evaluate the fitness of the population
        generator = GeneratorWithReturn(
            self.fandango.evaluator.evaluate_population(self.fandango.population)
        )
        solutions = list(generator)
        self.assertTrue(all(s.to_int() % 2 == 0) for s in solutions)

        self.assertEqual(len(solutions), len(set(solutions)))
        evaluation = generator.return_value
        self.assertEqual(len(evaluation), len(self.fandango.population))
        for ind, fitness, failing_trees in evaluation:
            self.assertIsInstance(fitness, float)
            self.assertGreaterEqual(fitness, 0.0)
            self.assertLessEqual(fitness, 1.0)
            self.assertIsInstance(failing_trees, list)
            for failing_tree in failing_trees:
                self.assertIsInstance(failing_tree, FailingTree)

        # Check that the population is valid
        for individual in self.fandango.population:
            self.assertTrue(self.fandango.grammar.parse(str(individual)))

    def test_select_elites(self):
        # Select the elites
        elites = self.fandango.evaluator.select_elites(
            self.fandango.evaluation,
            elitism_rate=self.fandango.elitism_rate,
            population_size=self.fandango.population_size,
        )

        self.assertEqual(
            len(elites), self.fandango.elitism_rate * self.fandango.population_size
        )

        # Check that the population is valid
        for individual in elites:
            self.assertTrue(self.fandango.grammar.parse(str(individual)))

    def test_selection(self):
        # Select the parents
        parent1, parent2 = self.fandango.evaluator.tournament_selection(
            self.fandango.evaluation,
            tournament_size=max(
                2, int(self.fandango.population_size * self.fandango.tournament_size)
            ),
        )

        # Check that the parents are in the population
        self.assertIn(parent1, self.fandango.population)
        self.assertIn(parent2, self.fandango.population)

        # Check that the parents are different
        self.assertNotEqual(parent1, parent2)

        # Check that the parents are of the correct type
        self.assertIsInstance(parent1, DerivationTree)
        self.assertIsInstance(parent2, DerivationTree)

        # Check that the population is valid
        for individual in [parent1, parent2]:
            self.assertTrue(self.fandango.grammar.parse(str(individual)))

    def test_crossover(self):
        # Select the parents
        tournament_size = max(
            2, int(self.fandango.population_size * self.fandango.tournament_size)
        )
        parent1, parent2 = self.fandango.evaluator.tournament_selection(
            self.fandango.evaluation,
            tournament_size,
        )

        # Perform crossover
        children = self.fandango.crossover_operator.crossover(
            self.fandango.grammar, parent1, parent2
        )

        # Check that the children are of the correct type
        for child in children:
            self.assertIsInstance(child, DerivationTree)

        # Check that the children are different
        self.assertNotEqual(children[0], children[1])

        # Check that the population is valid
        for individual in children:
            self.assertTrue(self.fandango.grammar.parse(str(individual)))

    def test_mutation(self):
        # Select the parents
        tournament_size = max(
            2, int(self.fandango.population_size * self.fandango.tournament_size)
        )
        parent1, parent2 = self.fandango.evaluator.tournament_selection(
            self.fandango.evaluation, tournament_size
        )

        children = self.fandango.crossover_operator.crossover(
            self.fandango.grammar, parent1, parent2
        )

        # Perform mutation
        gen1 = GeneratorWithReturn(
            self.fandango.mutation_method.mutate(
                children[0],
                self.fandango.grammar,
                self.fandango.evaluator.evaluate_individual,
            )
        )
        _solutions1 = list(gen1)
        mutant1 = gen1.return_value

        gen2 = GeneratorWithReturn(
            self.fandango.mutation_method.mutate(
                children[1],
                self.fandango.grammar,
                self.fandango.evaluator.evaluate_individual,
            )
        )
        _solutions2 = list(gen2)
        mutant2 = gen2.return_value

        # Check that the mutated children are of the correct type
        for child in [mutant1, mutant2]:
            self.assertIsInstance(child, DerivationTree)

        # Check that the mutated children are different
        self.assertNotEqual(mutant1, mutant2)

        # Check that the population is valid
        for individual in [mutant1, mutant2]:
            self.assertTrue(self.fandango.grammar.parse(str(individual)))

    def test_generate(self):
        # Run the evolution process
        solutions = list(self.fandango.generate(max_generations=20))

        # Check that the population has been updated
        self.assertIsNotNone(self.fandango.population)
        self.assertNotEqual(self.fandango.population, [])

        # Check that the population is valid
        for individual in self.fandango.population:
            self.assertTrue(self.fandango.grammar.parse(str(individual)))

        for individual in solutions:
            self.assertTrue(self.fandango.grammar.parse(str(individual)))


class DeterminismTests(unittest.TestCase):
    # fandango fuzz -f tests/resources/determinism.fan -n 100 --random-seed 1
    @staticmethod
    def get_solutions(
        specification_file,
        desired_solutions,
        random_seed,
    ) -> Generator[str, None, None]:
        with open(specification_file, "r") as file:
            grammar_int, constraints_int = parse(
                file, use_stdlib=False, use_cache=False
            )
        assert grammar_int is not None
        fandango = Fandango(
            grammar=grammar_int,
            constraints=constraints_int,
            random_seed=random_seed,
        )
        generator = itertools.islice(
            fandango.generate(max_generations=100), desired_solutions
        )
        for solution in generator:
            yield str(solution)

    def test_deterministic_solutions(self):
        gen1 = self.get_solutions(RESOURCES_ROOT / "determinism.fan", 30, 1)
        gen2 = self.get_solutions(RESOURCES_ROOT / "determinism.fan", 30, 1)

        self.assertListEqual(
            list(gen1), list(gen2), f"gen1: {list(gen1)}\ngen2: {list(gen2)}"
        )


class TargetedMutations(unittest.TestCase):
    # fandango fuzz -f tests/resources/digit_targeted_mutation.fan -n 1 --random-seed 1
    @staticmethod
    def get_solutions(
        specification_file,
        desired_solutions,
        random_seed,
    ):
        with open(specification_file, "r") as file:
            grammar_int, constraints_int = parse(
                file, use_stdlib=False, use_cache=False
            )
        assert grammar_int is not None
        fandango = Fandango(
            grammar=grammar_int,
            constraints=constraints_int,
            random_seed=random_seed,
        )
        solutions = []
        for solution in fandango.generate(max_generations=100):
            solutions.append(solution)
            if len(solutions) >= desired_solutions:
                break
        return [s.to_string() for s in solutions]

    def test_targeted_mutation_1(self):
        solutions = self.get_solutions(
            RESOURCES_ROOT / "digit_targeted_mutation.fan", 1, 1
        )
        self.assertListEqual(solutions, ["0123456789"])


if __name__ == "__main__":
    unittest.main()



from copy import deepcopy
from types import NoneType
from typing import Optional

from fandango.errors import FandangoValueError
from fandango.language.grammar import ParsingMode
from fandango.language.grammar.parser.column import Column
from fandango.language.grammar.node_visitors.node_visitor import NodeVisitor
from fandango.language.grammar.nodes.node import Node, NodeType
from fandango.language.grammar.nodes.non_terminal import NonTerminalNode
from fandango.language.grammar.nodes.terminal import TerminalNode
from fandango.language.grammar.nodes.repetition import Repetition
from fandango.language.grammar.nodes.alternative import Alternative
from fandango.language.grammar.nodes.concatenation import Concatenation
from fandango.language.grammar.nodes.repetition import Option
from fandango.language.grammar.nodes.repetition import Plus
from fandango.language.grammar.nodes.repetition import Star
from fandango.language.grammar.parser.parse_state import ParseState
from fandango.language.grammar.parser.parser_tree import ParserDerivationTree
from fandango.language.symbols import NonTerminal, Terminal
from fandango.language.tree import DerivationTree
from fandango.language.tree_value import TreeValue, TreeValueType


class IterativeParser(NodeVisitor):
    def __init__(
        self,
        grammar_rules: dict[NonTerminal, Node],
    ):
        self.implicit_start = NonTerminal("<*start*>")
        self.grammar_rules: dict[NonTerminal, Node] = grammar_rules
        self._rules: dict[NonTerminal, set[tuple[NonTerminal, frozenset]]] = {}
        self._implicit_rules: dict[NonTerminal, set[tuple[NonTerminal, frozenset]]] = {}
        self._context_rules: dict[
            NonTerminal, tuple[Node, tuple[NonTerminal, frozenset]]
        ] = dict()
        self._tmp_rules: dict[NonTerminal, set[tuple[NonTerminal, frozenset]]] = {}
        self._incomplete: set[DerivationTree] = set()
        self._nodes: dict[str, Node] = {}
        self._max_position = -1
        self.elapsed_time: float = 0.0
        self._process()
        self._table_idx = 0
        self._table: list[Column] = []
        self._parsing_mode = ParsingMode.COMPLETE
        self._bit_position = -1
        self._start: Optional[NonTerminal] = None
        self._first_consume = True
        self._hookin_parent: Optional[DerivationTree] = None
        self._prefix_word = None

    def _process(self):
        self._rules.clear()
        self._implicit_rules.clear()
        self._context_rules.clear()
        for nonterminal in self.grammar_rules:
            self.set_rule(nonterminal, self.visit(self.grammar_rules[nonterminal]))

        for nonterminal in self._implicit_rules:
            self._implicit_rules[nonterminal] = {
                tuple(a)  # type: ignore[misc]
                for a in self._implicit_rules[nonterminal]
            }

    def set_implicit_rule(
        self, rule: list[list[tuple[NonTerminal, frozenset]]]
    ) -> tuple[NonTerminal, frozenset]:
        nonterminal = NonTerminal(f"<*{len(self._implicit_rules)}*>")
        self._implicit_rules[nonterminal] = rule  # type: ignore[assignment]
        return (nonterminal, frozenset())

    def set_rule(
        self,
        nonterminal: NonTerminal,
        rule: list[list[tuple[NonTerminal, frozenset]]],
    ):
        self._rules[nonterminal] = {tuple(a) for a in rule}  # type: ignore[misc]

    def set_context_rule(
        self, node: Node, non_terminal: tuple[NonTerminal, frozenset]
    ) -> NonTerminal:
        nonterminal = NonTerminal(f"<*ctx_{len(self._context_rules)}*>")
        self._context_rules[nonterminal] = (node, non_terminal)
        return nonterminal

    def set_tmp_rule(
        self,
        rule: list[list[tuple[NonTerminal, frozenset]]],
        nonterminal: Optional[NonTerminal] = None,
    ):
        if nonterminal is None:
            nonterminal = NonTerminal(f"<*tmp_{len(self._tmp_rules)}*>")
        rule_id = str(nonterminal.value())[2:-2]
        self._tmp_rules[nonterminal] = {tuple(a) for a in rule}  # type: ignore[misc]
        return (nonterminal, frozenset()), rule_id

    def _clear_tmp(self):
        self._tmp_rules.clear()

    def default_result(self):
        return []

    def aggregate_results(self, aggregate, result):
        aggregate.extend(result)
        return aggregate

    def visitAlternative(self, node: Alternative):
        intermediate_nt = NonTerminal(f"<__{node.id}>")
        self._nodes[intermediate_nt.name()] = node
        result = self.visitChildren(node)
        self.set_rule(intermediate_nt, result)
        return [[(intermediate_nt, frozenset())]]

    def visitConcatenation(self, node: Concatenation):
        intermediate_nt = NonTerminal(f"<__{node.id}>")
        self._nodes[intermediate_nt.name()] = node
        result: list[list[tuple[NonTerminal, frozenset]]] = [[]]
        for child in node.children():
            to_add = self.visit(child)
            new_result = []
            for r in result:
                for a in to_add:
                    new_result.append(r + a)
            result = new_result
        self.set_rule(intermediate_nt, result)
        return [[(intermediate_nt, frozenset())]]

    def visitRepetition(
        self,
        node: Repetition,
        nt: Optional[tuple[NonTerminal, frozenset]] = None,
        tree: Optional[DerivationTree] = None,
    ):
        repetition_nt = NonTerminal(f"<__{node.id}>")
        self._nodes[repetition_nt.name()] = node
        is_context = node.bounds_constraint is not None

        if nt is None:
            alternatives = self.visit(node.node)
            nt = self.set_implicit_rule(alternatives)

            if is_context:
                i_nt = self.set_context_rule(node, nt)
                self.set_rule(repetition_nt, [[(i_nt, frozenset())]])
                return [[(repetition_nt, frozenset())]]

        prev = None
        if node.bounds_constraint is not None:
            assert tree is not None
            right_most_node = tree
            while len(right_most_node.children) != 0:
                right_most_node = right_most_node.children[-1]
            node_min, _ = node.bounds_constraint.min(right_most_node)
            node_max, _ = node.bounds_constraint.max(right_most_node)
        else:
            node_min = node.min
            node_max = node.max
        for rep in range(node_min, node_max):
            alts = [[nt]]
            if prev is not None:
                alts.append([nt, prev])
            if is_context:
                prev, rule_id = self.set_tmp_rule(alts)
            else:
                prev = self.set_implicit_rule(alts)
        alts = [node_min * [nt]]
        if prev is not None:
            alts.append(node_min * [nt] + [prev])
        if is_context:
            tmp_nt, rule_id = self.set_tmp_rule(alts)
            return [[tmp_nt]]
        min_nt = self.set_implicit_rule(alts)
        self.set_rule(repetition_nt, [[min_nt]])
        return [[(repetition_nt, frozenset())]]

    def visitStar(self, node: Star):
        intermediate_nt = NonTerminal(f"<__{node.id}>")
        self._nodes[intermediate_nt.name()] = node
        alternatives: list[list[tuple[NonTerminal, frozenset]]] = [[]]
        nt = self.set_implicit_rule(alternatives)
        for r in self.visit(node.node):
            alternatives.append(r + [nt])
        result = [[nt]]
        self.set_rule(intermediate_nt, result)
        return [[(intermediate_nt, frozenset())]]

    def visitPlus(self, node: Plus):
        intermediate_nt = NonTerminal(f"<__{node.id}>")
        self._nodes[intermediate_nt.name()] = node
        alternatives: list[list[tuple[NonTerminal, frozenset]]] = []
        nt = self.set_implicit_rule(alternatives)
        for r in self.visit(node.node):
            alternatives.append(r)
            alternatives.append(r + [nt])
        result = [[nt]]
        self.set_rule(intermediate_nt, result)
        return [[(intermediate_nt, frozenset())]]

    def visitOption(self, node: Option):
        intermediate_nt = NonTerminal(f"<__{node.id}>")
        self._nodes[intermediate_nt.name()] = node
        result: list[list[tuple[NonTerminal, frozenset]]] = [[]] + self.visit(node.node)
        self.set_rule(intermediate_nt, result)
        return [[(intermediate_nt, frozenset())]]

    def visitNonTerminalNode(self, node: NonTerminalNode):
        params = dict()
        if node.sender is not None:
            params["sender"] = node.sender
        if node.recipient is not None:
            params["recipient"] = node.recipient
        parameters = frozenset(params.items())
        return [[(node.symbol, parameters)]]

    def visitTerminalNode(self, node: TerminalNode):
        return [[(node.symbol, frozenset())]]

    def collapse(self, tree: Optional[DerivationTree]) -> Optional[DerivationTree]:
        if tree is None:
            return None
        if isinstance(tree.symbol, NonTerminal):
            if str(tree.symbol.value()).startswith("<__"):
                raise FandangoValueError(
                    "Can't collapse a tree with an implicit root node"
                )
        return self._collapse(tree)[0]

    def _collapse(self, tree: DerivationTree) -> list[DerivationTree]:
        reduced = []
        for child in tree.children:
            rec_reduced = self._collapse(child)
            reduced.extend(rec_reduced)

        if isinstance(tree.symbol, NonTerminal):
            if str(tree.symbol.value()).startswith("<__"):
                return reduced

        return [
            DerivationTree(
                tree.symbol,
                children=reduced,
                sources=tree.sources,
                read_only=tree.read_only,
                recipient=tree.recipient,
                sender=tree.sender,
                origin_repetitions=tree.origin_repetitions,
            )
        ]

    def can_continue(self):
        if len(self._table) <= 1:
            # Assume that an unstarted parse can continue
            return True
        table: list[Column] = list(self._table)
        table[self._table_idx] = deepcopy(table[self._table_idx])

        for state in table[-1]:
            if state.finished():
                self.complete(state, table, self._table_idx)

        return any(
            map(
                lambda state: state.is_incomplete or not state.finished(),
                table[self._table_idx],
            )
        )

    def predict(
        self,
        state: ParseState,
        table: list[Column],
        k: int,
        hookin_parent: Optional[DerivationTree] = None,
    ):
        symbol = state.dot
        assert symbol is not None
        assert isinstance(symbol, NonTerminal)
        if state.dot in self._rules:
            table[k].update(
                {
                    ParseState(state.dot, k, rule, 0)  # type: ignore[arg-type]
                    for rule in self._rules[symbol]
                }
            )
        elif state.dot in self._implicit_rules:
            table[k].update(
                {
                    ParseState(state.dot, k, rule, 0)  # type: ignore[arg-type]
                    for rule in self._implicit_rules[symbol]
                }
            )
        elif state.dot in self._tmp_rules:
            table[k].update(
                {
                    ParseState(state.dot, k, rule, 0)  # type: ignore[arg-type]
                    for rule in self._tmp_rules[symbol]
                }
            )
        elif state.dot in self._context_rules:
            node, nt = self._context_rules[symbol]
            self.predict_ctx_rule(state, table, k, node, nt, hookin_parent)

    def construct_incomplete_tree(
        self, state: ParseState, table: list[Column]
    ) -> DerivationTree:
        current_tree = ParserDerivationTree(state.nonterminal, state.children)
        current_state = state
        found_next_state = True
        while found_next_state:
            found_next_state = False
            for table_state in table[current_state.position].states:
                if table_state.dot == current_state.nonterminal:
                    current_state = table_state
                    found_next_state = True
                    break
            assert isinstance(current_tree.symbol, NonTerminal)
            if current_tree.symbol.name().startswith("<*"):
                current_tree = ParserDerivationTree(
                    current_state.nonterminal,
                    [*current_state.children, *current_tree.children],
                    **dict(current_state.dot_params or {}),
                )
            else:
                current_tree = ParserDerivationTree(
                    current_state.nonterminal,
                    [*current_state.children, current_tree],
                    **dict(current_state.dot_params or {}),
                )

        return current_tree.children[0]

    def predict_ctx_rule(
        self,
        state: ParseState,
        table: list[Column],
        k: int,
        node: Node,
        nt_rule,
        hookin_parent: Optional[DerivationTree] = None,
    ):
        if not isinstance(node, Repetition):
            raise FandangoValueError(f"Node {node} needs to be a Repetition")

        tree = self.construct_incomplete_tree(state, table)
        collapsed_tree = self.collapse(tree)
        assert collapsed_tree is not None
        tree = collapsed_tree
        if hookin_parent is not None:
            hookin_parent.set_children(hookin_parent.children + [tree])
        try:
            [[context_nt]] = self.visitRepetition(
                node, nt_rule, tree if hookin_parent is None else hookin_parent
            )
        except (ValueError, FandangoValueError):
            return
        finally:
            if hookin_parent is not None:
                hookin_parent.set_children(hookin_parent.children[:-1])
        new_symbols = []
        placed = False
        for symbol, dot_params in state.symbols:
            if symbol == state.dot and not placed:
                new_symbols.append(context_nt)
                placed = True
            else:
                new_symbols.append((symbol, dot_params))
        new_state = ParseState(
            state.nonterminal,
            state.position,
            tuple(new_symbols),
            state._dot,
            state.children,
            state.is_incomplete,
        )
        if state in table[k]:
            table[k].replace(state, new_state)
        self.predict(new_state, table, k)

    def scan_bit(
        self,
        state: ParseState,
        word: str | bytes,
        table: list[Column],
        k: int,
        w: int,
        bit_count: int,
    ) -> bool:
        """
        Scan a bit from the input `word`.
        `table` is the parse table (may be modified by this function).
        `table[k]` is the current column.
        `word[w]` is the current byte.
        `bit_count` is the current bit position (7-0).
        Return True if a bit was matched, False otherwise.
        """
        assert state.dot is not None
        assert state.dot.is_type(TreeValueType.TRAILING_BITS_ONLY)
        assert 0 <= bit_count <= 7

        if w >= len(word):
            return False

        # Get the highest bit. If `word` is bytes, word[w] is an integer.
        byte = ord(word[w]) if isinstance(word, str) else word[w]
        bit = (byte >> bit_count) & 1

        # LOGGER.debug(f"Checking {state.dot} against {bit}")
        match, match_length = state.dot.check(bit)
        if not match or match_length == 0:
            # LOGGER.debug(f"No match")
            return False

        # Found a match
        # LOGGER.debug(f"Found bit {bit}")
        next_state = state.next()
        tree = ParserDerivationTree(Terminal(bit))
        next_state.append_child(tree)
        # LOGGER.debug(f"Added tree {tree.to_string()!r} to state {next_state!r}")
        # Insert a new table entry with next state
        # This is necessary, as our initial table holds one entry
        # per input byte, yet needs to be expanded to hold the bits, too.

        # Add a new table row if the bit isn't already represented
        # by a row in the parsing table
        # if len(table) <= k + 1:
        #    table.insert(k + 1, Column())
        table[k + 1].add(next_state)

        # Save the maximum position reached, so we can report errors
        self._max_position = max(self._max_position, w)

        return True

    def scan_bytes(
        self,
        state: ParseState,
        word: str | bytes,
        table: list[Column],
        k: int,
        w: int,
    ) -> bool:
        """
        Scan a byte from th noe input `word`.
        `state` is the current parse state.
        `table` is the parse table.
        `table[k]` is the current column.
        `word[w]` is the current byte.
        Return True if a byte was matched, False otherwise.
        """

        assert state.dot is not None
        assert not (
            state.dot.is_type(TreeValueType.TRAILING_BITS_ONLY)
            or state.dot.is_type(TreeValueType.EMPTY)
        )
        assert not state.dot.is_regex

        # LOGGER.debug(f"Checking byte(s) {state.dot!r} at position {w:#06x} ({w}) {word[w:]!r}")

        check_word = word[w:]
        if state.is_incomplete:
            prev_terminal = state.children[-1]
            prev_val = prev_terminal.symbol.value()
            prev_val_raw: str | bytes
            if prev_val.is_type(TreeValueType.BYTES):
                prev_val_raw = bytes(prev_val)
                check_word = bytes(
                    TreeValue(prev_val_raw).append(TreeValue(check_word))
                )
            else:
                prev_val_raw = str(prev_val)
                check_word = str(TreeValue(prev_val_raw).append(TreeValue(check_word)))
        if state.dot.is_type(TreeValueType.BYTES):
            dot_len = len(bytes(state.dot.value()))
        else:
            dot_len = len(str(state.dot.value()))

        match, match_length = state.dot.check(check_word)
        table_idx_multiplier = 8

        if not match:
            if (w + dot_len - state.incomplete_idx) < len(word):
                return False
            match, match_length = state.dot.check(check_word, incomplete=True)
            if not match or match_length == 0:
                return False

            next_state = state.copy()
            next_state.incomplete_idx = match_length
            next_state.is_incomplete = True
            tree = ParserDerivationTree(Terminal(check_word[:match_length]))
            if state.is_incomplete:
                next_state.children[-1] = tree
            else:
                next_state.append_child(tree)
        else:
            next_state = state.next()
            next_state.is_incomplete = False
            next_state.incomplete_idx = 0
            tree = ParserDerivationTree(Terminal(check_word[:match_length]))
            if state.is_incomplete:
                next_state.children[-1] = tree
            else:
                next_state.append_child(tree)
        table[k + ((match_length - state.incomplete_idx) * table_idx_multiplier)].add(
            next_state
        )
        # LOGGER.debug(f"Next state: {next_state} at column {k + match_length}")
        self._max_position = max(self._max_position, w + match_length)

        return True

    def scan_regex(
        self,
        state: ParseState,
        word: str | bytes,
        table: list[Column],
        k: int,
        w: int,
        mode: ParsingMode,
    ) -> bool:
        """
        Scan a byte from the input `word`.
        `state` is the current parse state.
        `table` is the parse table.
        `table[k]` is the current column.
        `word[w]` is the current byte.
        Return (True, #bytes) if bytes were matched, (False, 0) otherwise.
        """

        assert state.dot is not None
        assert not (
            state.dot.is_type(TreeValueType.TRAILING_BITS_ONLY)
            or state.dot.is_type(TreeValueType.EMPTY)
        )
        assert state.dot.is_regex

        check_word = word[w:]
        prev_match_length = 0
        if state.is_incomplete:
            prev_terminal = state.children[-1]
            prev_val = prev_terminal.symbol.value()
            prev_val_raw: str | bytes
            if prev_val.is_type(TreeValueType.BYTES):
                prev_val_raw = bytes(prev_val)
                check_word = bytes(
                    TreeValue(prev_val_raw).append(TreeValue(check_word))
                )
            else:
                prev_val_raw = str(prev_val)
                check_word = str(TreeValue(prev_val_raw).append(TreeValue(check_word)))
            prev_match_length = len(prev_val_raw)

        table_idx_multiplier = 8
        match, match_length = state.dot.check(check_word)
        table_offset = match_length
        if match and match_length <= prev_match_length:
            match = False
            match_length = 0
        incomplete_match, incomplete_match_length = state.dot.check(
            check_word, incomplete=True
        )
        incomplete_table_offset = incomplete_match_length
        if not match:
            if not incomplete_match or (incomplete_match_length + w) < len(word):
                return False

        if match:
            next_state = state.next()
            next_state.is_incomplete = False
            next_state.incomplete_idx = 0
            tree = ParserDerivationTree(Terminal(check_word[:match_length]))
            if state.is_incomplete:
                next_state.children[-1] = tree
            else:
                next_state.append_child(tree)
            table[
                k + ((table_offset - state.incomplete_idx) * table_idx_multiplier)
            ].add(next_state)
        if incomplete_match:
            next_state = state.copy()
            next_state.is_incomplete = True
            next_state.incomplete_idx = incomplete_match_length
            tree = ParserDerivationTree(Terminal(check_word[:incomplete_match_length]))
            if state.is_incomplete:
                next_state.children[-1] = tree
            else:
                next_state.append_child(tree)
            table[
                k
                + (
                    (incomplete_table_offset - state.incomplete_idx)
                    * table_idx_multiplier
                )
            ].add(next_state)

        self._max_position = max(self._max_position, w + match_length)
        return True

    def _rec_to_derivation_tree(
        self,
        tree: DerivationTree,
        origin_repetitions: Optional[list[tuple[str, int, int]]] = None,
    ):
        if origin_repetitions is None:
            origin_repetitions = []

        rep_option = None
        is_controlflow_node = (
            isinstance(tree.symbol, NonTerminal) and tree.symbol.name() in self._nodes
        )
        if is_controlflow_node:
            nt = tree.symbol
            assert isinstance(nt, NonTerminal)
            node = self._nodes[nt.name()]
            if isinstance(node, Repetition):
                node.iteration += 1
                rep_option = (node.id, node.iteration, 0)

        children: list[DerivationTree] = []
        for child in tree.children:
            if is_controlflow_node:
                if rep_option is not None:
                    current_origin_repetitions = list(origin_repetitions) + [rep_option]
                    rep_option = (rep_option[0], rep_option[1], rep_option[2] + 1)
                else:
                    current_origin_repetitions = list(origin_repetitions)
            else:
                current_origin_repetitions = []

            children.append(
                self._rec_to_derivation_tree(child, current_origin_repetitions)
            )  # type: ignore[arg-type]

        return DerivationTree(
            tree.symbol,
            children,
            parent=tree.parent,
            sources=tree.sources,
            sender=tree.sender,
            recipient=tree.recipient,
            read_only=tree.read_only,
            origin_repetitions=origin_repetitions,
        )

    def to_derivation_tree(self, tree: "ParserDerivationTree"):
        if tree is None:
            return None
        return self._rec_to_derivation_tree(tree)  # type: ignore[arg-type]

    def complete(
        self,
        state: ParseState,
        table: list[Column],
        k: int,
        use_implicit: bool = False,
    ):
        for s in table[state.position].find_dot(state.nonterminal):
            dot_params = dict(s.dot_params)
            s = s.next()
            if state.nonterminal in self._rules:
                s.append_child(
                    ParserDerivationTree(
                        state.nonterminal, state.children, **dot_params
                    )
                )
            else:
                if use_implicit and state.nonterminal in self._implicit_rules:
                    s.append_child(
                        ParserDerivationTree(
                            NonTerminal(state.nonterminal.symbol),
                            state.children,
                            **s.dot_params,
                        )
                    )
                else:
                    s.extend_children(state.children)
            table[k].add(s)

    def place_repetition_shortcut(self, table: list[Column], k: int):
        col = table[k]
        states = col.states
        beginner_nts = [f"<__{NodeType.PLUS}:", f"<__{NodeType.STAR}:"]

        found_beginners = set()
        for state in states:
            if any(
                map(
                    lambda b: str(state.nonterminal.value()).startswith(b),
                    beginner_nts,
                )
            ):
                found_beginners.add(state.symbols[0][0])

        for beginner in found_beginners:
            current_col_state = None
            for state in states:
                if state.nonterminal == beginner:
                    if state.finished():
                        continue
                    if len(state.symbols) == 2 and state.dot == beginner:
                        current_col_state = state
                        break
            if current_col_state is None:
                continue
            new_state = current_col_state
            origin_states = table[current_col_state.position].find_dot(
                current_col_state.dot
            )
            if len(origin_states) != 1:
                continue
            origin_state = origin_states[0]
            while not any(
                map(
                    lambda b: str(origin_state.nonterminal.value()).startswith(b),
                    beginner_nts,
                )
            ):
                new_state = ParseState(
                    new_state.nonterminal,
                    origin_state.position,
                    new_state.symbols,
                    new_state._dot,
                    [*origin_state.children, *new_state.children],
                    new_state.is_incomplete,
                )
                origin_states = table[new_state.position].find_dot(new_state.dot)
                if len(origin_states) != 1:
                    new_state = None  # type: ignore[assignment]
                    break
                origin_state = origin_states[0]

            if new_state is not None:
                col.replace(current_col_state, new_state)

    def new_parse(
        self,
        start: str | NonTerminal = "<start>",
        mode: ParsingMode = ParsingMode.COMPLETE,
        hookin_parent: Optional[DerivationTree] = None,
        starter_bit=-1,
    ):
        if isinstance(start, str):
            start = NonTerminal(start)
        self._start = start
        self._table_idx = (7 - starter_bit) % 8
        self._table = []
        self._table.append(Column())
        self._first_consume = True
        self._incomplete.clear()
        self._max_position = -1
        self._parsing_mode = mode
        self._hookin_parent = deepcopy(hookin_parent)
        self._clear_tmp()

    def consume(self, char: str | bytes | int):
        for tree in self._consume(char):
            yield self.to_derivation_tree(tree)

    def _consume(self, char: str | bytes | int):
        assert self._start is not None, "Call new_parse() before consume()"
        if isinstance(char, int):
            char = bytes([char])
        word = char

        # If >= 0, indicates the next bit to be scanned (7-0)
        table = list(self._table)
        table.extend([Column() for _ in range(len(char) * 8)])
        # Add the start state at the first consume
        if self._first_consume:
            table[self._table_idx].add(
                ParseState(self.implicit_start, 0, ((self._start, frozenset()),))
            )
            self._first_consume = False
        curr_table_idx = self._table_idx
        curr_word_idx = 0

        while curr_table_idx < len(table):
            curr_bit_position = 7 - (curr_table_idx % 8)
            if curr_table_idx == len(table) - 1:
                self._table = list(table)
                if len(table) > 0:
                    self._table[-1] = deepcopy(table[-1])
                self._table_idx = curr_table_idx
            # True iff we have processed all characters
            # (or some bits of the last character)
            at_end = curr_word_idx >= len(word)
            for state in table[curr_table_idx]:
                if state.finished():
                    if state.nonterminal == self.implicit_start:
                        if at_end:
                            for child in state.children:
                                yield child

                    self.complete(state, table, curr_table_idx)
                else:
                    if not state.is_incomplete and state.next_symbol_is_nonterminal():
                        self.predict(state, table, curr_table_idx, self._hookin_parent)
                    else:
                        if state.dot.is_type(TreeValueType.TRAILING_BITS_ONLY):
                            # Scan a bit
                            match = self.scan_bit(
                                state,
                                word,
                                table,
                                curr_table_idx,
                                curr_word_idx,
                                curr_bit_position,
                            )
                        else:
                            if state.dot.is_regex:
                                match = self.scan_regex(
                                    state,
                                    word,
                                    table,
                                    curr_table_idx,
                                    curr_word_idx,
                                    self._parsing_mode,
                                )
                            else:
                                match = self.scan_bytes(
                                    state,
                                    word,
                                    table,
                                    curr_table_idx,
                                    curr_word_idx,
                                )

            if self._parsing_mode == ParsingMode.INCOMPLETE and at_end:
                for state in table[curr_table_idx]:
                    if len(state.children) == 0:
                        continue
                    if state.nonterminal == self.implicit_start:
                        for child in state.children:
                            if child not in self._incomplete:
                                self._incomplete.add(child)
                                yield child
                    self.complete(state, table, curr_table_idx)

            self.place_repetition_shortcut(table, curr_table_idx)
            curr_table_idx += 1
            if curr_table_idx % 8 == 0:
                curr_word_idx += 1

    def max_position(self):
        """Return the maximum position reached during parsing."""
        return self._max_position



# from https://github.com/amykyta3/speedy-antlr-example/blob/master/setup.py

import sys
import os
import platform
import fnmatch
import setuptools

target = platform.system().lower()
PLATFORMS = {"windows", "linux", "darwin", "cygwin"}
for known in PLATFORMS:
    if target.startswith(known):
        target = known


def run_setup(with_binary):
    if with_binary:
        extra_compile_args = {
            "windows": ["/DANTLR4CPP_STATIC", "/Zc:__cplusplus", "/std:c++17"],
            "linux": ["-std=c++17"],
            "darwin": ["-std=c++17", "-mmacosx-version-min=10.13"],
            "cygwin": ["-std=c++17"],
        }

        # Define an Extension object that describes the Antlr accelerator
        cpp_parser_dir = "src/fandango/language/cpp_parser"
        parser_ext = setuptools.Extension(
            # Extension name shall be at the same level as the sa_fandango_parser.py module
            name="fandango.language.parser.sa_fandango_cpp_parser",
            # Add the Antlr runtime source directory to the include search path
            include_dirs=[cpp_parser_dir + "/antlr4-cpp-runtime"],
            # Rather than listing each C++ file (Antlr has a lot!), discover them automatically
            sources=get_files(cpp_parser_dir, "*.cpp"),
            depends=get_files(cpp_parser_dir, "*.h"),
            extra_compile_args=extra_compile_args.get(target, []),
        )
        ext_modules = [parser_ext]
    else:
        ext_modules = []

    # Define a package
    setuptools.setup(
        # TODO: This should come from pyproject.toml
        name="fandango-fuzzer",
        version="0.9.0",
        description="Fandango produces myriads of high-quality random inputs to test programs, giving users unprecedented control over format and shape of the inputs.",
        packages=setuptools.find_packages("src"),
        package_dir={"": "src"},
        include_package_data=True,
        # python_requires='>=3.6.0',
        # install_requires=[
        #     "antlr4-python3-runtime >= 4.11, < 4.12",
        # ],
        ext_modules=ext_modules,
        cmdclass={"build_ext": ve_build_ext},
    )


# ===============================================================================
from setuptools.command.build_ext import build_ext
from setuptools.errors import CCompilerError, ExecError, PlatformError


def get_files(path, pattern):
    """
    Recursive file search that is compatible with python3.4 and older
    """
    matches = []
    for root, _, filenames in os.walk(path):
        for filename in fnmatch.filter(filenames, pattern):
            matches.append(os.path.join(root, filename))
    return matches


class BuildFailed(Exception):
    pass


class ve_build_ext(build_ext):
    """
    This class extends setuptools to fail with a common BuildFailed exception
    if a build fails
    """

    def run(self):
        try:
            build_ext.run(self)
        except PlatformError:
            raise BuildFailed()

    def build_extension(self, ext):
        try:
            build_ext.build_extension(self, ext)
        except (CCompilerError, ExecError, PlatformError):
            raise BuildFailed()
        except ValueError:
            # this can happen on Windows 64 bit, see Python issue 7511
            if "'path'" in str(sys.exc_info()[1]):  # works with Python 2 and 3
                raise BuildFailed()
            raise


# Detect if an alternate interpreter is being used
is_jython = "java" in sys.platform
is_pypy = hasattr(sys, "pypy_version_info")
skip_cpp = os.environ.get("FANDANGO_SKIP_CPP_PARSER", "") == "1"


# Force using fallback if using an alternate interpreter
using_fallback = is_jython or is_pypy or skip_cpp

if not using_fallback:
    try:
        run_setup(with_binary=True)
    except BuildFailed:
        if "FANDANGO_REQUIRE_BINARY_BUILD" in os.environ:
            raise
        else:
            using_fallback = True

if using_fallback:
    run_setup(with_binary=False)
